#!/usr/bin/env pytest

import os
import time
import pytest
import subprocess
from pathlib import Path

from ornithology import (
    action,
    write_file,
    format_script,
    Condor,
    ClusterState,
)

import htcondor2 as htcondor
import classad2 as classad

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

#
# If a plug-in fails partway through uploading a checkpoint, it doesn't
# necessarily delete the partial upload, so HTCondor has to track all
# upload attempts, and delete those which fail.
#
# We therefore need to test:
#   1.  That the shadow cleans up after an upload failure.
#   2.  That condor_preen cleans up after an upload failure, in case the
#       the shadow didn't try to clean up (because it failed) or because
#       its attempt failed for some other reason.
#   3.  That deleting the failed attemt does not delete any other attempt.
#   4.  That the starter does not attempt to download any failed
#       checkpoint.
#   5.  That a failure to checkpoint does not increase CommittedTime.
#
# We assume that if the starter downloads only succesful checkpoints that
# the job resumes appropriately; this is explicitly test elsewhere.
#
# The upload failure cases should specifically include:
#   a.  The first attempt failing, then the rest succeeding.
#   b.  The second attempt failing, and the rest succeeding.
#   c.  Alternate attempts succeeding and failing.
#   d.  All attempts failing.  (All attempts succeeding is checked
#       in other tests.)
#   e.  Alternate attempts suceeding and failing in pairs.
#
# Test (1) and (2) are mutually exclusive, but (3), (4), and (5) should
# be checked in both cases.  However, since (1) and (2) are mutually
# exclusive, we can just examine the state of the disk after the job
# completes to determine if they did the right thing.
#
# It would be easiest to implement (1) and (2) as different HTCondor
# configurations, implying that test cases (a) - (e) could otherwise
# be shared.  (They should be executed concurrently, unless (3) - (5)
# imply otherwise.)  In both (1) and (2), it would be easiest to
# submit the job with `on_exit_remove = false`, so that the schedd
# doesn't try to clean up when the job exits the queue.  (It should
# be possible to configure the schedd to run a useless clean-up
# script instead, but it seems easier to not run it at all.)
#
#   * Let's ignore (3) for now; it's wrong to delete _all_ failed checkpoints
#     only for performance reasons.  If that becomes a problem, we can
#     write a test static test for this later (that is, construct a fake set
#     of MANIFEST and FAILURE files in the right place, run the clean-up
#     script and see what's left).
#   * We should be able to check (4) after the fact by looking at the
#     transfer log.  (The URLs will all be distinct.)
#     Test (4) doesn't actually make sense if we've passed test (1),
#     so we could skip it in the shadow tests, but it won't hurt to
#     double-check, and if (1) fails but (4) succeeds, that probably
#     means we have an interesting problem....
#   * We should be able to check (5) by looking at the epoch log, because
#     all jobs evict themselves after their first upload failure.  For
#     simplicity, we'll only look at the test cases whose first checkpoint
#     fails, because then we know the CommittedTime should be 0.
#

TEST_CONFIGS = {
    "shadow": {
        "config": {
            # Prevent condor_preen from running.
            'PREEN':                    '/bin/false',
        },
        "script":                       'cleanup_locally_mounted_checkpoint',
    },
    "preen": {
        "config": {
        },
        "script":                       '/bin/false',
    },
}


# For simplicity, (a)-(d) could all attempt four checkpoints.  The lists would
# have to be updated, but we could also attempt eight checkpoints in all cases,
# if that would easier.
ONE_TEST_CASES = {
    "five": {
        "expected_checkpoints":     ["0000", "0001", "0002", "0004"],
        "checkpoint_count":         5,
    },
}

TEST_CASES = {
    "a": {
        "expected_checkpoints":     ["0001", "0002"],
        "checkpoint_count":         3,
    },
    "b": {
        "expected_checkpoints":     ["0000", "0002"],
        "checkpoint_count":         3,
    },
    "c1": {
        "expected_checkpoints":     ["0001", "0003"],
        "checkpoint_count":         4,
    },
    "c2": {
        "expected_checkpoints":     ["0000", "0002"],
        "checkpoint_count":         4,
    },
    "d": {
        "expected_checkpoints":     [],
        "checkpoint_count":         3,
    },
    "e1": {
        "expected_checkpoints":     ["0000", "0001", "0004", "0005"],
        "checkpoint_count":         6,
    },
    "e2": {
        "expected_checkpoints":     ["0000", "0001", "0004", "0005"],
        "checkpoint_count":         8,
    },
}


@action
def path_to_the_job_script(test_dir, the_condor):
    with the_condor.use_config():
        condor_libexec = htcondor.param["LIBEXEC"]
        condor_bin = htcondor.param["BIN"]

    script = f"""
    #!/usr/bin/python3

    import os
    import sys
    import time
    import subprocess
    from pathlib import Path

    os.environ['CONDOR_CONFIG'] = '{the_condor.config_file}'
    posix_test_dir = '{test_dir.as_posix()}'
    condor_libexec = '{condor_libexec}'
    condor_bin = '{condor_bin}'
    """ + """
    os.environ['PATH'] = condor_libexec + os.pathsep + os.environ.get('PATH', '') + os.pathsep + condor_bin

    count = 0

    print("Starting up...")
    if Path("saved-state-dir/FAILURE").exists():
        print("Found FAILURE file")

    try:
        with open("saved-state-dir/count", "r") as saved_state:
            count = int(saved_state.readline().strip())
            print(f"Starting from checkpoint at {count}.")
    except IOError:
        pass

    count +=1
    time.sleep(1)

    jobID = sys.argv[1]
    expected_checkpoints = sys.argv[2]
    num_checkpoints = int(sys.argv[3])

    if count > num_checkpoints:
        print("Completed counting.")
        sys.exit(0)

    path = Path("saved-state-dir")
    path.mkdir(exist_ok=True)

    with open(path / "count", "w") as saved_state:
        saved_state.write(f'{count}')

    fail_path = path / "FAILURE"
    checkpoint_number = count - 1
    the_count = f"{checkpoint_number:04}"
    if the_count in expected_checkpoints:
        # Sigh.  Python >= 3.8 only.
        # fail_path.unlink(missing_ok=True)
        try:
            fail_path.unlink()
        except FileNotFoundError:
            pass
    else:
        fail_path.touch()

    # Evict ourselves once just before the end.
    rv = subprocess.run(
        ["condor_q", "-jobads", os.environ['_CONDOR_JOB_AD'], "-af", "NumJobStarts"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        timeout=20
    )
    if rv.returncode != 0:
        print(f'Failed to find NumJobStarts in .job.ad')
        print(f'{rv.stdout}')
        print(f'{rv.stderr}')
        sys.exit(1)
    epoch = int(rv.stdout.decode('utf-8').strip())

    print(f"epoch {epoch}")
    should_evict = False
    if epoch == 0 and not the_count in expected_checkpoints:
        should_evict = True
    the_previous_count = f"{(checkpoint_number-1):04}"
    if epoch == 1 and the_previous_count in expected_checkpoints:
        should_evict = True

    if should_evict:
        print("evicting myself...", flush=True)
        rv = subprocess.run(
            ["condor_vacate_job", jobID],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            timeout=20
        )
        if rv.returncode != 0:
            print(f'Failed to vacate myself after second checkpoint:')
            print(f'{rv.stdout}')
            print(f'{rv.stderr}')
            sys.exit(1)

        # Don't exit before we've been vacated.
        time.sleep(60)

        # Signal a failure if we weren't vacated.
        print("Failed to evict, aborting.")
        sys.exit(1)

    print("... checkpointing")
    sys.exit(85)
    """

    path = test_dir / "job_script.py"
    write_file(path, format_script(script))

    return path


@action(params={name: name for name in TEST_CONFIGS})
def the_config_tuple(request):
    return (request.param, TEST_CONFIGS[request.param])


@action
def the_config_name(the_config_tuple):
    return the_config_tuple[0]


@action
def the_config(the_config_tuple):
    return the_config_tuple[1]


@action
def the_condor(test_dir, the_config, the_config_name):
    local_dir = test_dir / f"condor.{the_config_name}"

    with Condor(
        local_dir=local_dir,
        config={
            'ENABLE_URL_TRANSFERS':             'TRUE',
            'JOB_EPOCH_HISTORY':                '$(LOG)/EpochLog',
            'SCHEDD_INTERVAL':                  '2',
            'SHADOW.FILE_TRANSFER_STATS_LOG':   '$(LOG)/shadow_transfer_history',
            'SHADOW_DEBUG':                     'D_CATEGORY D_SUB_SECOND D_STATUS D_TEST',
            ** the_config['config'],
        }
    ) as the_condor:
        (local_dir / "checkpoint-destination-mapfile").write_text(
            f"*   local://example.vo/example.fs   {the_config['script']},-prefix,\\0,-path,{test_dir.as_posix()}\n"
        )
        yield the_condor


@action
def the_job_description(test_dir, the_config_name, path_to_the_job_script, path_to_the_plugin):
    return {
        # Setting this means that all terminal job states leave the job
        # in the queue until it's removed, which is convenient.
        "LeaveJobInQueue":              "true",

        "executable":                   path_to_the_job_script.as_posix(),
        "transfer_executable":          "true",
        "should_transfer_files":        "true",
        "when_to_transfer_output":      "ON_EXIT",
        "stream_output":                "true",
        "stream_error":                 "true",

        "log":                          test_dir / f"{the_config_name}_test_job_$(CLUSTER).log",
        "output":                       test_dir / f"{the_config_name}_test_job_$(CLUSTER).out",
        "error":                        test_dir / f"{the_config_name}_test_job_$(CLUSTER).err",
        "want_io_proxy":                "True",

        "checkpoint_exit_code":         "85",
        "transfer_checkpoint_files":    "saved-state-dir",
        "transfer_output_files":        "",

        "checkpoint_destination":       'local://example.vo/example.fs',
        "transfer_plugins":             f'local={path_to_the_plugin}',
    }


@action
def the_job_handles(the_condor, the_job_description):
    job_handles = {}
    for name, test_case in TEST_CASES.items():
        expected_checkpoint_list = ",".join(test_case['expected_checkpoints'])
        arguments = f"$(CLUSTER).$(PROCESS) {expected_checkpoint_list} {test_case['checkpoint_count']}"
        complete_job_description = {
            'arguments': arguments,
            ** the_job_description,
        }

        job_handle = the_condor.submit(
            description=complete_job_description,
            count=1,
        )

        job_handles[name] = job_handle

    yield job_handles


@action(params={name: name for name in TEST_CASES})
def the_job_tuple(request, the_job_handles):
    return (request.param, the_job_handles[request.param])


@action
def the_job_name(the_job_tuple):
    return the_job_tuple[0]


@action
def the_expected_checkpoints(the_job_name):
    return TEST_CASES[the_job_name]["expected_checkpoints"]


@action
def the_job_handle(the_job_tuple):
    return the_job_tuple[1]


@action
def the_completed_job(the_condor, the_job_handle):
    # We now set SCHEDD_INTERVAL to a small enough value in the test suite
    # that we don't have explicitly call for a reschedule to run the test
    # in a timely fashion.
    assert the_job_handle.wait(
        timeout=300,
        condition=ClusterState.all_complete,
        fail_condition=ClusterState.any_held,
    )

    return the_job_handle


@action
def the_completed_job_ad(the_condor, the_completed_job):
    schedd = the_condor.get_local_schedd()
    constraint = f'ClusterID == {the_completed_job.clusterid} && ProcID == 0'
    result = schedd.query(
            constraint=constraint,
            projection=[
                "CheckpointDestination",
                "GlobalJobID",
                "ClusterID",
                "ProcID",
                "Owner",
            ],
    )

    return result[0]


@action
def the_target_directory(test_dir, the_completed_job_ad):
    globalJobID = the_completed_job_ad["GlobalJobID"].replace('#', '_')
    return test_dir / globalJobID


@action
def the_transfers(test_dir, the_completed_job_ad, the_condor):
    # Find the transfer history file.
    with the_condor.use_config():
        FILE_TRANSFER_STATS_LOG = htcondor.param["FILE_TRANSFER_STATS_LOG"]

    # Each ClassAd in the transfer history file should have 'JobClusterID'
    # and 'JobProcID' attribute.
    clusterID = the_completed_job_ad["ClusterID"]
    procID = the_completed_job_ad["ProcID"]

    # Find each transfer ad for this job.
    transfers = []
    globalJobID = the_completed_job_ad["GlobalJobID"].replace('#', '_')
    with open(FILE_TRANSFER_STATS_LOG, "r") as transfer_history:
            #
            # This is _amazingly_ stupid.  Why don't any of the condor tools
            # actually handle this?  Why is there no Python API that does?
            #

            # Eat the first "separator" line.
            line = transfer_history.readline()
            while True:
                buffer = ''

                while True:
                    line = transfer_history.readline()
                    if line == "":
                        # Because Python has no eof() function.
                        break
                    elif line != "***\n":
                        buffer += line
                    else:
                        ad = classad.parseOne(buffer)

                        # We only care about the job we're looking at.
                        if ad['JobClusterID'] != clusterID:
                            break
                        if ad['JobProcID'] != procID:
                            break

                        # What I want to generate here was a list of all
                        # the checkpoint numbers downloaded by the starter.
                        #
                        # The original idea was to look at the URLs downloaded
                        # by the starter and parse them to find the checkpoint
                        # numbers, and report the list of those.
                        #
                        # For some reason, the file transfer object doesn't
                        # actually log downloaded URLs.  It also doesn't log
                        # CEDAR uploads, and it logs uploads from the starter
                        # to a URL as CEDAR downloads.
                        #
                        # The upshot of this disaster is that even adding
                        # which subsystem logged to the transfer doesn't
                        # let me figure out what actually happened.
                        #
                        # Instead, I'm going to look for the starter's records
                        # of downloaded MANIFEST files, and declare those to be
                        # a true indication of what actually happened.
                        #

                        # However, we can set SHADOW.FILE_TRANSFER_STATS_LOG
                        # and keep the shadow events out of the default
                        # transfer history file without (further) hacks.
                        # if ad['Subsystem'] != "STARTER":
                        #   break
                        if ad['TransferProtocol'] != "cedar":
                            break
                        # Just in case we ever fix any of the above.
                        if ad['TransferType'] != 'download':
                            break

                        # The ./ seems redundant, but it's always there.
                        prefix = "./_condor_checkpoint_MANIFEST."
                        transferFileName = ad['TransferFileName']
                        if transferFileName.startswith(prefix):
                            # Only for Python >= 3.9 :(
                            # suffix = transferFileName.removeprefix(prefix)
                            suffix = transferFileName[len(prefix):]
                            transfers.append(suffix[0:4])

                        break

                if line == "" and buffer == "":
                    break

    # print(f"TRANSFERS: {transfers}")
    return transfers


@action
def the_removed_job(the_condor, the_completed_job):
    the_condor.run_command(['condor_qedit', the_completed_job.clusterid, 'LeaveJobInQueue', 'False'])
    the_completed_job.remove()

    while True:
        # This makes me sad.
        time.sleep(1)
        history_ads = the_condor.get_local_schedd().history(f"ClusterID == {the_completed_job.clusterid}", ["ClusterID"])
        if sum(1 for _ in history_ads) > 0:
            # Job has hit the history file, now we can continue
            return the_completed_job
        # but assert if the job is held, so we don't go into an infinite loop
        assert not the_completed_job.state.any_held()

    return False


@action
def the_preen_directory(the_removed_job, the_target_directory, the_condor, test_dir, the_config_name):
    # To aid in debugging, don't run condor_preen after a shadow test.
    if the_config_name != "preen":
        pytest.skip("preen-only test")
        return the_target_directory

    logger.debug("logging all ads")
    all_job_ads = the_condor.query(projection=["ClusterID", "ProcID"])
    for ad in all_job_ads:
        logger.debug(ad)

    with the_condor.use_config():
        SPOOL = htcondor.param["SPOOL"]

        local_dir = Path(htcondor.param["LOCAL_DIR"])
        (local_dir / "checkpoint-destination-mapfile").write_text(
            f"*   local://example.vo/example.fs   cleanup_locally_mounted_checkpoint,-prefix,\\0,-path,{test_dir.as_posix()}\n"
        )

        preen_env = {
            ** os.environ,
            '_CONDOR_SPOOL': SPOOL,
            '_CONDOR_TOOL_DEBUG': 'D_TEST D_SUB_SECOND D_CATEGORY',
        }
        rv = subprocess.run( ['condor_preen', '-d'],
            # Crass empiricism.
            env=preen_env, timeout=120,
            stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        logger.debug(rv.stdout)
        assert(rv.returncode == 0)

    return the_target_directory


@action
def the_epoch_log(the_condor, the_completed_job_ad, the_removed_job):
    schedd = the_condor.get_local_schedd()
    globalJobID = the_completed_job_ad["GlobalJobID"]
    constraint = f'globalJobID == "{globalJobID}"'
    return [ad for ad in schedd.jobEpochHistory(constraint, [])]


class TestPartialUploads:

    # Test (1).
    def test_s_removes_failed_checkpoints(self,
        the_config_name,
        the_target_directory,
        the_expected_checkpoints
    ):
        if the_config_name != "shadow":
            pytest.skip("shadow-only test")
            return

        # ... and every expected checkpoint should be on disk.
        for checkpoint in the_expected_checkpoints[-2:]:
            path = the_target_directory / checkpoint
            assert(path.exists())

        # If we don't expect any checkpoints, then the test below
        # will fail because the_target_directory doesn't exist.
        if len(the_expected_checkpoints) == 0:
            assert not the_target_directory.exists()
            return

        # Everything on disk should be an expected checkpoint...
        for path in the_target_directory.iterdir():
            assert(path.is_dir())
            assert(path.name in the_expected_checkpoints)


    # Test (2).
    def test_p_removes_failed_checkpoints(self,
        the_condor,
        the_config_name,
        the_preen_directory,
        the_completed_job_ad
    ):
        if the_config_name != "preen":
            pytest.skip("preen-only test")
            return

        assert(not the_preen_directory.exists())

        # In addition to cleaning up the checkpoint destination, we should
        # make certain that we clean up our own SPOOL directory; this test
        # triggers a case the test_checkpoint_destination does not, so we
        # need to check here, too.  (Where condor_preen and the schedd both
        # try to clean up checkpoints whose MANIFEST files don't exist,
        # causing condor_manifest to think that the clean-up failed; the
        # tool now cleans up the .job.ad file (under checkpoint-cleanup/)
        # and its containing directory if it's the only file in there.
        with the_condor.use_config():
            SPOOL = htcondor.param["SPOOL"]
            cleanup_dir = Path(SPOOL) / "checkpoint-cleanup" / the_completed_job_ad["Owner"] / f"cluster{the_completed_job_ad['ClusterID']}.proc{the_completed_job_ad['ProcID']}.subproc0"
            assert(not cleanup_dir.exists())


    # Test (4).
    def test_download_only_successful_checkpoints(self,
        the_transfers,
        the_expected_checkpoints
    ):
        assert (len(the_expected_checkpoints) != 0) == (len(the_transfers) != 0)

        for transfer in the_transfers:
            assert(transfer in the_expected_checkpoints)


    # Test (5).
    def test_committed_time(self,
        the_job_name,
        the_epoch_log,
        the_expected_checkpoints
    ):
        if '0000' in the_expected_checkpoints:
            pytest.skip('for simplicitly, check only for 0 committed time')
            return

        first_epoch_ad = list(reversed(the_epoch_log))[0]
        assert first_epoch_ad['CommittedTime'] == 0

        # for ad in reversed(the_epoch_log):
        #   print(f"{ad['ClusterID']}.{ad['ProcID']}: {ad['CommittedTime']}")


# Copied from test_checkpoint_destination, and modified to fail to upload
# files named FAILURE.
@action
def path_to_the_plugin(test_dir):
    path_to_the_plugin = test_dir / "local.py"

    contents = format_script(
    f"""
        #!/usr/bin/python3

        import os
        import sys
        import time
        import shutil

        from urllib.parse import urlparse
        from pathlib import Path

        import classad2 as classad

        DEFAULT_TIMEOUT = 30
        PLUGIN_VERSION = '1.0.0'

        EXIT_SUCCESS = 0
        EXIT_FAILURE = 1
        EXIT_AUTHENTICATION_REFRESH = 2

        test_dir = '{test_dir.as_posix()}'

""" + """

        def print_help(stream = sys.stderr):
            help_msg = '''Usage: {0} -infile <input-filename> -outfile <output-filename>
               {0} -classad

        Options:
          -classad                    Print a ClassAd containing the capablities of this
                                      file transfer plugin.
          -infile <input-filename>    Input ClassAd file
          -outfile <output-filename>  Output ClassAd file
          -upload                     Indicates this transfer is an upload (default is
                                      download)
        '''
            stream.write(help_msg.format(sys.argv[0]))

        def print_capabilities():
            capabilities = {
                 'MultipleFileSupport': True,
                 'PluginType': 'FileTransfer',
                 'SupportedMethods': 'local',
                 'Version': PLUGIN_VERSION,
            }
            sys.stdout.write(classad.ClassAd(capabilities).printOld())

        def parse_args():

            # The only argument lists that are acceptable are
            # <this> -classad
            # <this> -infile <input-filename> -outfile <output-filename>
            # <this> -outfile <output-filename> -infile <input-filename>
            if not len(sys.argv) in [2, 5, 6]:
                print_help()
                sys.exit(EXIT_FAILURE)

            # If -classad, print the capabilities of the plugin and exit early
            if (len(sys.argv) == 2) and (sys.argv[1] == '-classad'):
                print_capabilities()
                sys.exit(EXIT_SUCCESS)

            # If -upload, set is_upload to True and remove it from the args list
            is_upload = False
            if '-upload' in sys.argv[1:]:
                is_upload = True
                sys.argv.remove('-upload')

            # -infile and -outfile must be in the first and third position
            if not (
                    ('-infile' in sys.argv[1:]) and
                    ('-outfile' in sys.argv[1:]) and
                    (sys.argv[1] in ['-infile', '-outfile']) and
                    (sys.argv[3] in ['-infile', '-outfile']) and
                    (len(sys.argv) == 5)):
                print_help()
                sys.exit(1)
            infile = None
            outfile = None
            try:
                for i, arg in enumerate(sys.argv):
                    if i == 0:
                        continue
                    elif arg == '-infile':
                        infile = sys.argv[i+1]
                    elif arg == '-outfile':
                        outfile = sys.argv[i+1]
            except IndexError:
                print_help()
                sys.exit(EXIT_FAILURE)

            return {'infile': infile, 'outfile': outfile, 'upload': is_upload}

        def format_error(error):
            return '{0}: {1}'.format(type(error).__name__, str(error))

        def get_error_dict(error, url = ''):
            error_string = format_error(error)
            error_dict = {
                'TransferSuccess': False,
                'TransferError': error_string,
                'TransferUrl': url,
            }

            return error_dict

        class LocalPlugin:

            # Extract whatever information we want from the url provided.
            # In this example, convert the example://path/to/file url to a
            # path in the file system (ie. /path/to/file)
            def parse_url(self, url):
                url_path = url[(url.find("://") + 3):]
                return url_path

            def download_file(self, url, local_file_path):
                start_time = time.time()

                remote_file_path = self.parse_url(url)
                remote_file_path = remote_file_path.replace( "example.vo/example.fs", test_dir )

                # print(f"DEBUG: {start_time} download {remote_file_path} -> {local_file_path}")
                file_size = Path(remote_file_path).stat().st_size
                Path(local_file_path).parent.mkdir(parents=True,exist_ok=True)
                shutil.copy(remote_file_path, local_file_path)

                end_time = time.time()

                # Get transfer statistics
                transfer_stats = {
                    'TransferSuccess': True,
                    'TransferProtocol': 'example',
                    'TransferType': 'upload',
                    'TransferFileName': local_file_path,
                    'TransferFileBytes': file_size,
                    'TransferTotalBytes': file_size,
                    'TransferStartTime': int(start_time),
                    'TransferEndTime': int(end_time),
                    'ConnectionTimeSeconds': end_time - start_time,
                    'TransferUrl': url,
                }

                return transfer_stats

            def upload_file(self, url, local_file_path):

                start_time = time.time()

                remote_file_path = self.parse_url(url)
                remote_file_path = remote_file_path.replace( "example.vo/example.fs", test_dir )

                # print(f"DEBUG: {start_time} upload {remote_file_path} <- {local_file_path}")
                file_size = Path(local_file_path).stat().st_size
                Path(remote_file_path).parent.mkdir(parents=True,exist_ok=True)

                success = True
                if local_file_path.endswith("FAILURE"):
                    success = False
                if success:
                    shutil.copy(local_file_path, remote_file_path)

                end_time = time.time()

                # Get transfer statistics
                transfer_stats = {
                    'TransferSuccess': success,
                    'TransferProtocol': 'example',
                    'TransferType': 'upload',
                    'TransferFileName': local_file_path,
                    'TransferFileBytes': file_size,
                    'TransferTotalBytes': file_size,
                    'TransferStartTime': int(start_time),
                    'TransferEndTime': int(end_time),
                    'ConnectionTimeSeconds': end_time - start_time,
                    'TransferUrl': url,
                }
                if not success:
                    transfer_stats["TransferError"] = "b/c you said so"

                return transfer_stats


        if __name__ == '__main__':

            # Start by parsing input arguments
            try:
                args = parse_args()
            except Exception:
                sys.exit(EXIT_FAILURE)

            local_plugin = LocalPlugin()

            # Parse in the classads stored in the input file.
            # Each ad represents a single file to be transferred.
            try:
                infile_ads = classad.parseAds(open(args['infile'], 'r'))
            except Exception as err:
                try:
                    with open(args['outfile'], 'w') as outfile:
                        outfile_dict = get_error_dict(err)
                        outfile.write(str(classad.ClassAd(outfile_dict)))
                except Exception:
                    pass
                sys.exit(EXIT_FAILURE)

            # Now iterate over the list of classads and perform the transfers.
            try:
                with open(args['outfile'], 'w') as outfile:
                    for ad in infile_ads:
                        try:
                            if not args['upload']:
                                outfile_dict = local_plugin.download_file(ad['Url'], ad['LocalFileName'])
                            else:
                                outfile_dict = local_plugin.upload_file(ad['Url'], ad['LocalFileName'])

                            outfile.write(str(classad.ClassAd(outfile_dict)))

                        except Exception as err:
                            try:
                                outfile_dict = get_error_dict(err, url = ad['Url'])
                                outfile.write(str(classad.ClassAd(outfile_dict)))
                            except Exception:
                                pass
                            sys.exit(EXIT_FAILURE)

            except Exception:
                sys.exit(EXIT_FAILURE)

            if not args['upload']:
                mutator = Path('mutate_checkpoint.py')
                if mutator.exists():
                    os.system(f"./{mutator}")
    """
    )

    write_file(path_to_the_plugin, contents)
    return path_to_the_plugin



import os

from docutils.parsers.rst import Directive
from docutils.parsers.rst import directives
from docutils import nodes

from htc_helpers import warn

HIDE_ENV_VAR = "SPHINX_HIDE_HIDDEN_SECTIONS"

class HiddenSection(Directive):
    """Custom directive to hide documentation in offical build but not local"""
    optional_arguments = 1 # Allow a specified version history to no longer be hidden
    has_content = True # Allow additional content (to hide)
    option_spec = {"comment" : directives.flag} # Allow :comment: option to always hide

    # Note returns list[nodes.Node]
    def run(self) -> list:
        """Execute this directive"""
        if "comment" in self.options:
            return []

        # Get Sphinx BuildEnvironment for later information
        env = self.state.document.settings.env

        # Check the current release versrion against labeled release version
        if len(self.arguments) > 0:
            RELEASE_VERSION = tuple(int(x) for x in env.config.release.split("."))
            version = tuple(int(x) for x in self.arguments[0].split("."))
            if version <= RELEASE_VERSION:
                warn(f"{env.docname}:{self.lineno} | Hidden section labeled for v{self.arguments[0]} release... currently v{env.config.release}")

        info = list()

        # Check if we should actually parse the contents for display or hide
        if os.getenv(HIDE_ENV_VAR, "False") not in ["True", "true", "1"]:
            text = '\n'.join(self.content)
            node = nodes.container(text)
            self.add_name(node)
            self.state.nested_parse(self.content, self.content_offset, node, match_titles=True)
            info.append(node)

        return info

def setup(app):
    """Setup sphinx to know of this directive"""
    app.add_directive('hidden', HiddenSection)



#!/usr/bin/env pytest


#
# Submit a job whose plug-in takes long enough to trigger the callback
# mechanism, then check to make sure that the D_TEST output from doing
# so appears in the starter log.
#
# Submit a job whose plug-in sleeps for long enough to trigger
# STARTER_FILE_XFER_STALL_TIMEOUT (which must be less than
# MAX_FILE_TRANSFER_PLUGIN_LIFETIME), then check to verify that the
# callback was succesful in preventing the starter from EXCEPT()ing.
#


import logging

from ornithology import *

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

import htcondor2
from pathlib import Path


@action
def the_condor(test_dir):
    local_dir = test_dir / "the_condor"
    with Condor(
        local_dir=local_dir,
        config={
            "STARTER_DEBUG": "D_CATEGORY D_SUB_SECOND D_TEST D_PID",
            "STARTER_XFER_LAST_ALIVE_DELAY": 5,
            "STARTER_XFER_LAST_ALIVE_INTERVAL": 5,
            "MY_POPEN_KEEPALIVE_INTERVAL": 15,
            "STARTER_FILE_XFER_STALL_TIMEOUT": 25,
            "MAX_FILE_TRANSFER_PLUGIN_LIFETIME": 75,
        }
    ) as the_condor:
        yield the_condor


@action
def the_completed_job(the_condor):
    job_description = {
        "universe":                 "vanilla",
        "shell":                    "sleep 1",

        "log":                      "the_job.log",

        "should_transfer_files":    True,
        "transfer_input_files":     "debug://sleep/50",
    }

    job_handle = the_condor.submit(
        description=job_description,
        count=1
    )

    assert job_handle.wait(
        timeout=100,
        condition=ClusterState.all_complete,
    )

    return job_handle


# Stolen from `test_cif.py`.
def count_starter_log_lines(the_condor, the_phrase):
    with the_condor.use_config():
        LOG = htcondor2.param["LOG"]

    count = 0
    for log in Path(LOG).iterdir():
        if log.name.startswith("StarterLog.slot1_"):
            for line in log.read_text().split("\n"):
                if the_phrase in line:
                    count = count + 1

    return count


class TestHungPlugin:

    # If we got to this test at all, that means the job finished (and
    # finished the first time it ran), which probably means things are
    # all OK, but let's be explicit.
    def test_callback_fires(self, the_condor, the_completed_job):
        num_multi_callbacks = count_starter_log_lines(
            the_condor, "wait_for_output() callback (multi)"
        )
        assert num_multi_callbacks >= 1

        num_starter_exceptions = count_starter_log_lines(
            the_condor, "EXCEPT"
        )
        assert num_starter_exceptions == 0



#!/usr/bin/env pytest

# Test that OCU (Owner Capacity Unit) creation and job matching works:
# 1. Stand up a personal condor with default partitionable slots
# 2. Submit an OCUWanted job (this creates the SubmitterData the schedd
#    needs before it will negotiate for OCU claims)
# 3. Create an OCU via the Python API
# 4. Wait for the OCU-wanting job to start running on the OCU slot
# 5. While the OCU job is still running (so the OCU slot persists),
#    submit a non-OCU job and verify it does NOT run on the OCU slot

import logging
import time
import getpass

import htcondor2 as htcondor
import classad2 as classad

from ornithology import *

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@standup
def condor(test_dir):
	with Condor(
		test_dir / "condor",
		config={
			"NUM_CPUS": "4",
			"MEMORY": "4096",
			"DISK": "4096000",
			"NEGOTIATOR_CYCLE_DELAY": "2",
			"NEGOTIATOR_INTERVAL": "2",
		},
	) as condor:
		yield condor


@action
def ocu_job_handle(condor, test_dir, path_to_sleep):
	"""Submit a job that wants an OCU slot.  This must happen before
	the OCU is created so the schedd has a SubmitterData record for
	our user, which is required for the OCU to be negotiated.
	The job sleeps long enough to remain running while we submit
	and verify the non-OCU job."""
	handle = condor.submit(
		description={
			"executable": path_to_sleep,
			"arguments": "300",
			"log": (test_dir / "ocu_job.log").as_posix(),
			"leave_in_queue": True,
			"should_transfer_files": False,
			"Request_Cpus": 1,
			"Request_Memory": 1,
			"Request_Disk": 1,
			"+OCUWanted": "true",
		},
		count=1,
	)

	yield handle
	handle.remove()


@action
def the_ocu(condor, ocu_job_handle):
	"""Create an OCU after the job is submitted so the SubmitterData exists."""
	schedd = condor.get_local_schedd()

	request = classad.ClassAd()
	request["Owner"] = getpass.getuser()
	request["RequestCpus"] = 1
	request["RequestMemory"] = 1
	request["RequestDisk"] = 1

	result = schedd.create_ocu(request)
	logger.debug(f"OCU create result: {result}")
	assert result["Result"] == 0, f"Failed to create OCU: {result.get('ErrorString', 'unknown')}"

	yield result

	# Cleanup: remove the OCU
	try:
		schedd.remove_ocu(int(result["OCUId"]))
	except Exception:
		pass


@action
def ocu_slot_name(condor, ocu_job_handle, the_ocu):
	"""Wait for the OCU job to start running, then extract the slot name.
	We keep the OCU job running so the OCU slot persists while we test
	that a non-OCU job avoids it."""
	ocu_job_handle.wait(
		condition=ClusterState.all_running,
		fail_condition=ClusterState.any_held,
		verbose=True,
		timeout=120,
	)

	for ad in ocu_job_handle.query():
		return get_remote_host(ad)


@action
def non_ocu_job_handle(condor, test_dir, path_to_sleep, ocu_slot_name):
	"""Submit a normal job (no OCUWanted) and wait for completion."""
	handle = condor.submit(
		description={
			"executable": path_to_sleep,
			"arguments": "0",
			"log": (test_dir / "non_ocu_job.log").as_posix(),
			"leave_in_queue": True,
			"should_transfer_files": False,
			"Request_Cpus": 1,
			"Request_Memory": 1,
			"Request_Disk": 1,
		},
		count=1,
	)

	handle.wait(
		condition=ClusterState.all_complete,
		fail_condition=ClusterState.any_held,
		verbose=True,
		timeout=120,
	)

	yield handle
	handle.remove()


@action
def non_ocu_job_ad(non_ocu_job_handle):
	"""Get the completed job ad for the non-OCU job."""
	for ad in non_ocu_job_handle.query():
		return ad


def get_remote_host(job_ad):
	"""Extract the host the job ran on, handling the race between
	RemoteHost and LastRemoteHost."""
	if "LastRemoteHost" in job_ad:
		return job_ad["LastRemoteHost"]
	return job_ad["RemoteHost"]


class TestOCUBasic:
	def test_ocu_created(self, the_ocu):
		"""Verify the OCU was created successfully."""
		assert the_ocu["Result"] == 0
		assert "OCUId" in the_ocu
		assert "OCUName" in the_ocu

	def test_ocu_job_ran_on_ocu_slot(self, ocu_slot_name):
		"""Verify a job with OCUWanted=true was matched to a slot."""
		assert ocu_slot_name is not None

	def test_non_ocu_job_avoided_ocu_slot(self, non_ocu_job_ad, ocu_slot_name):
		"""Verify a normal job did NOT run on the OCU slot."""
		host = get_remote_host(non_ocu_job_ad)
		assert host != ocu_slot_name, (
			f"Non-OCU job ran on OCU slot {ocu_slot_name}, but should not have"
		)



import sys
import enum


class QueryOpt(enum.IntFlag):
    """
    An enumeration of the flags that may be sent to the *condor_schedd*
    during a query to alter its behavior.

    .. attribute:: Default

        Queries should use default behaviors, and return jobs for all users.

    .. attribute:: AutoCluster

        Instead of returning job ads, return an ad per auto-cluster.

    .. attribute:: GroupBy

        Instead of returning job ads, return an ad for each unique combination
        of values for the attributes in the projection.  Similar to
        ``AutoCluster``, but using the projection as the significant attributes
        for auto-clustering.

    .. attribute:: DefaultMyJobsOnly

        Queries should use all default behaviors, and return jobs only for the
        current user.

    .. attribute:: SummaryOnly

        Instead of returning job ads, return only the final summary ad.

    .. attribute:: IncludeClusterAd

        Query should return raw cluster ads as well as job ads if the cluster
        ads match the query constraint.
    """

    Default = 0x00
    AutoCluster = 0x01
    GroupBy = 0x02
    DefaultMyJobsOnly = 0x04
    SummaryOnly = 0x08
    IncludeClusterAd = 0x10
    IncludeJobsetAds = 0x20



# Copyright 2022 HTCondor Team, Computer Sciences Department,
# University of Wisconsin-Madison, WI.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import json
import pprint

from pathlib import Path

try:
    import opensearchpy
    from opensearchpy import VERSION as OS_VERSION
    _OS_MODULE_FOUND = True
except ModuleNotFoundError as err:
    _OS_MODULE_FOUND = False
    _OS_MODULE_NOT_FOUND_ERROR = err


from adstash.utils import get_host_port
from adstash.interfaces.elasticsearch import ElasticsearchInterface


if _OS_MODULE_FOUND and (OS_VERSION < (1,0,0) or OS_VERSION >= (3,0,0)):
    logging.warning(f"Unsupported Opensearch Python library {OS_VERSION}, proceeding anyway...")


class OpenSearchInterface(ElasticsearchInterface):


    def __init__(
            self,
            host="localhost",
            port="9200",
            url_prefix="",
            username=None,
            password=None,
            use_https=False,
            ca_certs=None,
            timeout=60,
            **kwargs
            ):
        if not _OS_MODULE_FOUND:  # raise module not found error if missing
            raise _OS_MODULE_NOT_FOUND_ERROR
        self.host, self.port = get_host_port(host, port)
        self.url_prefix = url_prefix or ""
        self.username = username
        self.password = password
        self.use_https = use_https
        self.ca_certs = ca_certs
        self.timeout = timeout
        self.handle = None


    def get_handle(self):
        if self.handle is not None:
            return self.handle

        client_options = {}
        client_options["hosts"] = [{
            "host": self.host,
            "port": self.port,
            "url_prefix": self.url_prefix,
            "use_ssl": self.use_https,
        }]

        if (self.username is None) and (self.password is None):
            pass  # anonymous auth
        elif (self.username is None) != (self.password is None):
            logging.warning("Only one of username and password have been defined, attempting anonymous connection to OpenSearch")
        else:  # basic auth
            auth_tuple = (self.username, self.password,)
            client_options["http_auth"] = auth_tuple

        if self.ca_certs is not None:
            client_options["ca_certs"] = self.ca_certs
        if self.use_https:
            client_options["verify_certs"] = True

        client_options["timeout"] = self.timeout

        self.handle = opensearchpy.OpenSearch(**client_options)
        return self.handle


    def setup_index(self, index, log_mappings=True, log_dir=Path.cwd(), **kwargs):
        client = self.get_handle()

        # check if index is an alias, and get the active index if so
        if client.indices.exists_alias(name=index):
            index = self.get_active_index(index)

        mappings = self.make_mappings(**kwargs)
        settings = self.make_settings(**kwargs)

        if not client.indices.exists(index=index):  # push new index if doesn't exist
            logging.info(f"Creating new index {index}.")
            if OS_VERSION >= (2,0,0):
                body = {
                    "mappings": mappings,
                    "settings": settings,
                }
                client.indices.create(index=index, body=json.dumps(body))
            else:
                client.indices.create(index=index, mappings=mappings, settings=settings)
            if log_mappings and log_dir:
                mappings_file = log_dir / "condor_adstash_opensearch_last_mappings.json"
                logging.debug(f"Writing new mappings to {mappings_file}.")
                json.dump(mappings, open(mappings_file, "w"), indent=2)
            return

        # otherwise check existing index for missing mappings properties
        update_mappings = False
        updated_mappings = {}
        existing_mappings = self.get_mappings(index)
        for outer_key in mappings:
            if outer_key not in existing_mappings:  # add anything missing
                updated_mappings[outer_key] = mappings[outer_key]
                update_mappings = True
            elif isinstance(mappings[outer_key], dict):  # update missing keys in any existing dicts
                missing_inner_keys = set(mappings[outer_key]) - set(existing_mappings[outer_key])
                if len(missing_inner_keys) > 0:
                    updated_mappings[outer_key] = {}
                    for inner_key in missing_inner_keys:
                        updated_mappings[outer_key][inner_key] = mappings[outer_key][inner_key]
                    update_mappings = True
        if update_mappings:
            logging.info(f"Updated mappings for index {index}")
            logging.debug(f"{pprint.pformat(updated_mappings)}")
            if OS_VERSION >= (2,0,0):
                body = {"mappings": updated_mappings}
                client.indices.put_mapping(index=index, body=body)
            else:
                client.indices.put_mapping(index=index, **updated_mappings)
            if log_mappings and log_dir:
                mappings_file = log_dir / "condor_adstash_opensearch_last_mappings.json"
                logging.debug(f"Writing updated mappings to {mappings_file}.")
                json.dump(updated_mappings, open(mappings_file, "w"), indent=2)



import os
import sys

from docutils import nodes
from docutils.parsers.rst import Directive

from htc_helpers import make_headerlink_node, warn

TICKETS_ANCHORED = set()

def make_link_node(rawtext, app, type, slug, options):
    """Create a link to a JIRA ticket.

    :param rawtext: Text being replaced with link node.
    :param app: Sphinx application context
    :param type: Link type (issue, changeset, etc.)
    :param slug: ID of the ticket to link to
    :param options: Options dictionary passed to role func.
    """
    base = "https://opensciencegrid.atlassian.net/browse/HTCONDOR-"
    ref = base + slug
    # set_classes(options)
    node = nodes.reference(rawtext, "(HTCONDOR-" + slug + ")", refuri=ref, **options)
    return node

def make_version_anchor(inliner, options, tag):
    global TICKETS_ANCHORED
    tag = f"{tag}".replace(" ", "-").replace(",", "-")

    if tag in TICKETS_ANCHORED:
        entry_no = inliner.document.settings.env.new_serialno(tag)
        tag += f"-{entry_no}"
    else:
        TICKETS_ANCHORED.add(tag)

    parent = inliner.parent
    if "ids" in parent:
        parent["ids"].append(tag)
    else:
        parent["ids"] = [tag]

    return make_headerlink_node(tag, options)

def vers_hist_anchor_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    tag = "version-history" if text.lower() == "default" else text
    return [make_version_anchor(inliner, options, tag)], []

def ticket_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    nodes = list()
    for ticket_id_str in text.split(","):
        try:
            ticket_id = int(ticket_id_str)
            app = inliner.document.settings.env.app
            node = make_link_node(rawtext, app, 'issue', str(ticket_id), options)
            nodes.append(node)
        except ValueError as e:
            docname = inliner.document.settings.env.docname
            warn(f"{docname}:{lineno} | Failed to link ticket #{ticket_id_str}: {e}")

    if len(nodes) > 0:
        nodes.append(make_version_anchor(inliner, options, text))

    return nodes, []

def setup(app):
    app.add_role("jira", ticket_role)
    app.add_role("hist-anchor", vers_hist_anchor_role)




#!/usr/bin/env pytest

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

import time
from ornithology import *

import htcondor2


@action
def the_condor(test_dir):
    local_dir = test_dir / "condor"

    with Condor(
        local_dir=local_dir,
        config={
            "DAEMON_LIST":              "TEST_STDF_TIMER_D COLLECTOR",
            "TEST_STDF_TIMER_D":        "$(LIBEXEC)/test_stdf_timer_d",
            # Ornithology needs this, for some reason.
            "TEST_STDF_TIMER_D_LOG":    f"$(LOG)/TestStdFTimerDLog",
            "DC_DAEMON_LIST":           "+ TEST_STDF_TIMER_D",
        },
    ) as condor:
        yield condor


@action
def the_daemon_log(the_condor):
    with the_condor.use_config():
        libexec = htcondor2.param['LIBEXEC']
    for i in range(1,10):
        the_log = the_condor._get_daemon_log('TEST_STDF_TIMER_D')
        for line in the_log.open().read():
            if "EXITING WITH STATUS" in line.message:
                return the_log
        time.sleep(12)
    assert False


class TestDCStdFunction:

    def test_timer(self, the_condor, the_daemon_log):
        # I'm not at all sure I can just re-open the daemon log.
        the_log = the_condor._get_daemon_log('TEST_STDF_TIMER_D')

        for line in the_log.open().read():
            if "std::function timer fired." == line.message:
                return
            if "c function timer fired." == line.message:
                assert False
        assert False




#!/usr/bin/env pytest

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

from ornithology import (
    Condor,
    action,
    ClusterState,
    format_script,
    write_file,
    JobStatus,
)

import os
from pathlib import Path
import shutil

import htcondor2


def try_shutil_chown( * p, ** v ):
    try:
        shutil.chown( * p, ** v )
    except PermissionError as pe:
        logger.debug(pe)
    except LookupError as le:
        logger.debug(le)


# ---- test_one_cif_job -------------------------------------------------------

@action
def user_dir(test_dir):
    the_user_dir = test_dir / "user"
    the_user_dir.mkdir(exist_ok=True)
    try_shutil_chown( the_user_dir.as_posix(), user='tlmiller', group='tlmiller' )
    return the_user_dir


@action
def the_lock_dir(test_dir):
    return test_dir / "lock.d"


@action
def the_condor(test_dir, the_lock_dir):
    local_dir = test_dir / "condor"
    cred_dir = local_dir / "cred.d"
    # I don't think Ornithology knows about this one yet.
    cred_dir.mkdir(parents=True, exist_ok=True)

    with Condor(
        submit_user='tlmiller',
        condor_user='condor',
        local_dir=local_dir,
        config={
            "STARTER_DEBUG":    "D_CATEGORY D_SUB_SECOND D_PID D_ACCOUNTANT",
            "SHADOW_DEBUG":     "D_CATEGORY D_SUB_SECOND D_PID D_TEST",
            "LOCK":             the_lock_dir.as_posix(),
            "DAEMON_LIST":      "$(DAEMON_LIST) CREDD",
            "SEC_CREDENTIAL_DIRECTORY_OAUTH": cred_dir.as_posix(),
            # Test for long sinful strings.
            "HOST_ALIAS":       "abcdefghijlmnopqrstuvwxyzabcdefghijlmnopqrstuvwxyz"
                              + "abcdefghijlmnopqrstuvwxyzabcdefghijlmnopqrstuvwxyz"
                              + "abcdefghijlmnopqrstuvwxyzabcdefghijlmnopqrstuvwxyz"
                              + "abcdefghijlmnopqrstuvwxyzabcdefghijlmnopqrstuvwxyz"
                              + "abcdefghijlmnopqrstuvwxyzabcdefghijlmnopqrstuvwxyz"
                              + "abcdefghijlmnopqrstuvwxyzabcdefghijlmnopqrstuvwxyz",
        },
    ) as the_condor:
        yield the_condor


@action
def completed_cif_job(the_condor, path_to_sleep, user_dir):
    os.chdir(user_dir.as_posix())

    d = user_dir / "d"
    d.mkdir()
    (d / "input1.txt").write_text("input")
    (d / "input4.txt").write_text(" A\n")

    e = user_dir / "d" / "e"
    e.mkdir()
    (e / "input2.txt").write_text("II")

    f = user_dir / "d" / "f"
    f.mkdir()
    (f / "input5.txt").write_text(" input\n")

    (user_dir / "input3.txt").write_text("input line 3\n" );

    # Make sure that credential propogation works, too.
    credential_path = user_dir / "the_credential"
    credential_path.write_text("fake credential information")
    cp = the_condor.run_command(
        ['htcondor', 'credential', 'add', 'oauth2', credential_path.as_posix()],
        timeout=60,
        echo=True,
        as_user='tlmiller',
    )
    assert cp.returncode == 0

    job_description = {
        "universe":                 "vanilla",

        "shell":                    "cat ${_CONDOR_CREDS}/the_credential.use 1>&2; cat d/input1.txt d/input4.txt d/e/input2.txt d/f/input5.txt input3.txt; sleep 5",
        "transfer_executable":      False,
        "should_transfer_files":    True,

        "log":                      "cif_job.log",
        "output":                   "cif_job.output",
        "error":                    "cif_job.error",

        "request_cpus":             1,
        "request_memory":           1,

        "MY._x_common_input_catalogs":      '"my_common_files"',
        "MY._x_catalog_my_common_files":    '"d, null://create-epoch-entry"',

        "transfer_input_files":     "input3.txt",
        "use_oauth_services":       "the_credential",

        "leave_in_queue":           True,
    }

    job_handle = the_condor.submit(
        description=job_description,
        count=1,
    )

    assert job_handle.wait(
        timeout=200,
        condition=ClusterState.all_complete,
    )

    return job_handle


# ---- test_many_cif_jobs -----------------------------------------------------


@action
def the_big_lock_dir(test_dir):
    return test_dir / "big.lock.d"


@action
def the_big_condor(test_dir, the_big_lock_dir):
    local_dir = test_dir / "big-condor"

    with Condor(
        submit_user='tlmiller',
        condor_user='condor',
        local_dir=local_dir,
        config={
            "STARTER_DEBUG":    "D_CATEGORY D_SUB_SECOND D_PID D_ACCOUNTANT",
            "SHADOW_DEBUG":     "D_CATEGORY D_SUB_SECOND D_PID D_TEST",
            "LOCK":             the_big_lock_dir.as_posix(),
            "NUM_CPUS":         4,
            "STARTER_ALLOW_RUNAS_OWNER":    False,
            "SLOT1_1_USER":     "glump",
            "SLOT1_2_USER":     "sorcy",
            "SLOT1_3_USER":     "kittie",
            "SLOT1_4_USER":     "jrandom",
            "STARTER_NESTED_SCRATCH":   False,
        },
    ) as the_condor:
        yield the_condor


@action
def cif_jobs_script(test_dir):
    script_path = test_dir / "big_job_script"

    script_text = format_script(
    """
        #!/bin/bash

        cat big_input1.txt big_input2.txt big_input3.txt

        while true; do
            sleep 1
            if [ -e $1 ]; then
                exit 0
            fi
        done

        exit 1
    """
    )
    write_file(script_path, script_text)
    return script_path


@action
def completed_cif_jobs(the_big_condor, user_dir, cif_jobs_script):
    os.chdir(user_dir.as_posix())
    (user_dir / "big_input1.txt").write_text("input A\n")
    (user_dir / "big_input2.txt").write_text("II input\n");
    (user_dir / "big_input3.txt").write_text("input line 3\n" );

    job_description = {
        "universe":                 "vanilla",

        "executable":               cif_jobs_script.as_posix(),
        "arguments":                (user_dir / "kill-many-$(ClusterID).$(ProcID)").as_posix(),
        "transfer_executable":      True,
        "should_transfer_files":    True,

        "log":                      "cif_job.log.$(CLUSTER)",
        "output":                   "cif_job.output.$(CLUSTER).$(PROCESS)",
        "error":                    "cif_job.error.$(CLUSTER).$(PROCESS)",

        "request_cpus":             1,
        "request_memory":           1,

        # Force a delay; the constant is to ensure that we would print out
        # the waiting message at least twice (if the delay remains at five
        # seconds each time).
        "MY._x_common_input_catalogs":      '"my_common_files"',
        "MY._x_catalog_my_common_files":    '"big_input1.txt, big_input2.txt, debug://sleep/15"',

        "transfer_input_files":     "big_input3.txt",

        "leave_in_queue":           True,
        "hold":                     True,
    }

    # Submit eight jobs on hold.
    job_handle = the_big_condor.submit(
        description=job_description,
        count=8,
    )

    # Release four of the jobs.
    jobIDs = [ f"{job_handle.clusterid}.{x}" for x in range(0,4) ]
    the_big_condor.act( htcondor2.JobAction.Release, jobIDs )

    # Wait for them to start.
    assert job_handle.wait(
        timeout=200,
        condition=ClusterState.running_exactly(4),
        fail_condition=ClusterState.any_terminal
    )

    # Touch the kill files on the first four jobs, causing them to finish.
    for i in range(0,4):
        (user_dir / f"kill-many-{job_handle.clusterid}.{i}").touch(exist_ok=True)

    # Wait for them to finish.
    assert job_handle.wait(
        timeout=200,
        condition=lambda self: self.status_exactly(4, JobStatus.COMPLETED),
        fail_condition=ClusterState.any_held
    )

    # This test used to overlap releasing some jobs while others were still
    # running, to make sure that "delayed" concurrent re-use worked properly.
    # We could stagger the start of either (or both) halves of the jobs,
    # but we're also using this test to verify obseverability, which means
    # we need to check for a specific number of times what jobs wait for
    # common file transfer to happen.  That leads to a race, because we don't
    # know that the common file transfer has completed by the time wait()
    # returns (which only knows about shadow start-up).
    #
    # Of course, we intend to add common files transfer events to the job
    # event log later on in this PR, so maybe using those (after checking
    # them after the fact in test_one_cif_job) would work well.

    # Release the other four jobs.
    jobIDs = [ f"{job_handle.clusterid}.{x}" for x in range(4,8) ]
    the_big_condor.act( htcondor2.JobAction.Release, jobIDs )

    # Wait for the last four jobs to start.
    assert job_handle.wait(
        timeout=200,
        condition=ClusterState.none_idle,
    )

    # Touch the kill files on the last four jobs, causing them to finish.
    for i in range(4,8):
        (user_dir / f"kill-many-{job_handle.clusterid}.{i}").touch(exist_ok=True)

    # Wait for the last four jobs to finish.
    assert job_handle.wait(
        timeout=200,
        condition=ClusterState.all_complete,
    )

    return job_handle


# ---- test_multi_cif_jobs ----------------------------------------------------


@action
def the_multi_lock_dir(test_dir):
    return test_dir / "multi.lock.d"


@action
def the_multi_condor(test_dir, the_multi_lock_dir):
    local_dir = test_dir / "multi-condor"

    with Condor(
        submit_user='tlmiller',
        condor_user='condor',
        local_dir=local_dir,
        config={
            "STARTER_DEBUG":    "D_CATEGORY D_SUB_SECOND D_PID D_ACCOUNTANT",
            "SHADOW_DEBUG":     "D_CATEGORY D_SUB_SECOND D_PID D_TEST",
            "LOCK":             the_multi_lock_dir.as_posix(),
            "NUM_CPUS":         4,
            "STARTER_NESTED_SCRATCH":   True,
        },
    ) as the_condor:
        yield the_condor


@action
def multi_job_script(test_dir):
    script_path = test_dir / "multi_job_script"

    script_text = format_script(
    """
        #!/bin/bash

        killfile=$1; shift
        cat "$@"

        while true; do
            sleep 1
            if [ -e $killfile ]; then
                exit 0
            fi
        done

        exit 1
    """
    )
    write_file(script_path, script_text)
    return script_path


@action
def completed_multi_jobs(the_multi_condor, user_dir, multi_job_script):
    os.chdir(user_dir.as_posix())

    #
    # In this test, we're validating that different CIFs stay different.
    #

    (user_dir / "multi_input1.txt").write_text("input A\n")
    (user_dir / "multi_input2.txt").write_text("II input\n");
    (user_dir / "multi_input3.txt").write_text("input line 3\n" );

    kill_file = (user_dir / "kill-multi-$(ClusterID).$(ProcID)").as_posix()
    job_description_a = {
        "universe":                 "vanilla",

        "executable":               multi_job_script.as_posix(),
        "arguments":
            f'{kill_file} multi_input1.txt multi_input2.txt multi_input3.txt',
        "transfer_executable":      True,
        "should_transfer_files":    True,

        "log":                      "multi_job.log.$(CLUSTER)",
        "output":                   "multi_job.output.$(CLUSTER).$(PROCESS)",
        "error":                    "multi_job.error.$(CLUSTER).$(PROCESS)",

        "request_cpus":             1,
        "request_memory":           1,

        "MY._x_common_input_catalogs":      '"my_common_files"',
        "MY._x_catalog_my_common_files":    '"multi_input1.txt, multi_input2.txt"',

        "transfer_input_files":     "multi_input3.txt",

        "leave_in_queue":           True,
        "hold":                     True,
    }


    (user_dir / "multi_input4.txt").write_text("-- input A\n")
    (user_dir / "multi_input5.txt").write_text("-- II input\n");
    (user_dir / "multi_input6.txt").write_text("-- input line 3\n" );

    job_description_b = { ** job_description_a,
        "arguments":
            f'{kill_file} multi_input4.txt multi_input5.txt multi_input6.txt',
        "MY._x_common_input_catalogs":      '"my_common_files"',
        "MY._x_catalog_my_common_files":    '"multi_input4.txt, multi_input5.txt"',
        "transfer_input_files":     "multi_input6.txt",
    }


    # Submit eight jobs on hold.
    job_handle_a = the_multi_condor.submit(
        description=job_description_a,
        count=4,
    )

    job_handle_b = the_multi_condor.submit(
        description=job_description_b,
        count=4,
    )


    # Release four of the jobs.
    jobIDs = [ f"{job_handle_a.clusterid}.{x}" for x in range(0,2) ]
    the_multi_condor.act( htcondor2.JobAction.Release, jobIDs )
    jobIDs = [ f"{job_handle_b.clusterid}.{x}" for x in range(0,2) ]
    the_multi_condor.act( htcondor2.JobAction.Release, jobIDs )


    # Wait for them to start.
    assert job_handle_a.wait(
        timeout=200,
        condition=ClusterState.running_exactly(2),
        fail_condition=ClusterState.any_terminal
    )
    assert job_handle_b.wait(
        timeout=200,
        condition=ClusterState.running_exactly(2),
        fail_condition=ClusterState.any_terminal
    )

    # Touch the kill files on the first four jobs, causing them to finish.
    for i in range(0,2):
        (user_dir / f"kill-multi-{job_handle_a.clusterid}.{i}").touch(exist_ok=True)
    for i in range(0,2):
        (user_dir / f"kill-multi-{job_handle_b.clusterid}.{i}").touch(exist_ok=True)

    # Wait for the first four jobs to finish.
    # Due to a bug in Ornithology, jobs submitted on hold but not yet
    # released are considered IDLE, which is... not helpful.
    assert job_handle_a.wait(
        timeout=200,
        condition=lambda s: s.all_status(JobStatus.IDLE, JobStatus.COMPLETED),
    )
    assert job_handle_b.wait(
        timeout=200,
        condition=lambda s: s.all_status(JobStatus.IDLE, JobStatus.COMPLETED),
    )

    # Release the other four jobs.
    jobIDs = [ f"{job_handle_a.clusterid}.{x}" for x in range(2,4) ]
    the_multi_condor.act( htcondor2.JobAction.Release, jobIDs )
    jobIDs = [ f"{job_handle_b.clusterid}.{x}" for x in range(2,4) ]
    the_multi_condor.act( htcondor2.JobAction.Release, jobIDs )

    # Wait for the last four jobs to start.
    assert job_handle_a.wait(
        timeout=200,
        condition=ClusterState.none_idle,
    )
    assert job_handle_b.wait(
        timeout=200,
        condition=ClusterState.none_idle,
    )

    # Touch the kill files on the last four jobs, causing them to finish.
    for i in range(2,4):
        (user_dir / f"kill-multi-{job_handle_a.clusterid}.{i}").touch(exist_ok=True)
    for i in range(2,4):
        (user_dir / f"kill-multi-{job_handle_b.clusterid}.{i}").touch(exist_ok=True)

    # Wait for the last four jobs to finish.
    assert job_handle_a.wait(
        timeout=200,
        condition=ClusterState.all_complete,
    )
    assert job_handle_b.wait(
        timeout=200,
        condition=ClusterState.all_complete,
    )

    return job_handle_a, job_handle_b


# ---- assertion helpers ------------------------------------------------------


def output_is_as_expected(completed_cif_job, the_expected_output):
    ads = completed_cif_job.query( projection=['Out'] )

    for ad in ads:
        out = Path(ad['Out'])
        text = out.read_text()
        assert text == the_expected_output


def error_is_as_expected(completed_cif_job, the_expected_error):
    ads = completed_cif_job.query( projection=['Err'] )

    for ad in ads:
        err = Path(ad['Err'])
        text = err.read_text()
        assert text == the_expected_error


def count_shadow_log_lines(the_condor, the_phrase):
    shadow_log = the_condor.shadow_log.open()
    lines = list(filter(
        lambda line: the_phrase in line,
        shadow_log.read(),
    ))
    return len(lines)


def shadow_log_is_as_expected(the_condor, count, cf_xfers, cf_waits):
    staging_commands_sent = count_shadow_log_lines(
        the_condor, "StageCommonFiles"
    )
    assert staging_commands_sent == count

    successful_staging_commands = count_shadow_log_lines(
        the_condor, "Staging successful"
    )
    assert successful_staging_commands == count

    keyfile_touches = count_shadow_log_lines(
        the_condor, "Producer elected"
    )
    assert keyfile_touches == count

    job_evictions = count_shadow_log_lines(
        the_condor, "is being evicted from"
    )
    assert job_evictions == 0

    common_transfer_begins = count_shadow_log_lines(
        the_condor, "Starting common files transfer."
    )
    assert common_transfer_begins == cf_xfers

    common_transfer_ends = count_shadow_log_lines(
        the_condor, "Finished common files transfer: success."
    )
    assert common_transfer_ends == cf_xfers

    if cf_waits is not None:
        common_transfer_waits = count_shadow_log_lines(
            the_condor, "Waiting for common files to be transferred"
        )
        assert common_transfer_waits == cf_waits


def lock_dir_is_clean(the_lock_dir):
    syndicate_dir = the_lock_dir / "syndicate"

    files = list(syndicate_dir.iterdir())
    assert len(files) == 0


def epoch_log_has_common_ad(the_condor):
    with the_condor.use_config():
        JOB_EPOCH_HISTORY = htcondor2.param["JOB_EPOCH_HISTORY"]

    epoch_log = Path(JOB_EPOCH_HISTORY)
    for line in epoch_log.read_text().split('\n'):
        if '*** COMMON ClusterId' in line:
            return True

    return False


def count_starter_log_lines(the_condor, the_phrase):
    with the_condor.use_config():
        LOG = htcondor2.param["LOG"]

    count = 0
    for log in Path(LOG).iterdir():
        if log.name.startswith("StarterLog.slot1_"):
            for line in log.read_text().split("\n"):
                if the_phrase in line:
                    count = count + 1

    return count


def starter_log_is_as_expected(the_condor, cf_xfers, cf_waits):
    common_transfer_begins = count_starter_log_lines(
        the_condor, "Starting common files transfer."
    )
    assert common_transfer_begins == cf_xfers

    common_transfer_ends = count_starter_log_lines(
        the_condor, "Finished common files transfer: success."
    )
    assert common_transfer_ends == cf_xfers

    common_transfer_waits = count_starter_log_lines(
        the_condor, "Waiting for common files to be transferred"
    )
    assert common_transfer_waits == cf_waits


# ---- the tests --------------------------------------------------------------


class TestCIF:

    def test_one_cif_job(self, the_lock_dir, the_condor, completed_cif_job):
        output_is_as_expected(
            completed_cif_job, "input A\nII input\ninput line 3\n"
        )
        shadow_log_is_as_expected(the_condor, 1, 1, 0)
        lock_dir_is_clean(the_lock_dir)
        error_is_as_expected(
            completed_cif_job, "fake credential information"
        )

        starter_log_is_as_expected(the_condor, 1, 0)
        assert epoch_log_has_common_ad(the_condor)


    def test_many_cif_jobs(self, the_big_lock_dir, the_big_condor, completed_cif_jobs):
        output_is_as_expected(
            completed_cif_jobs, "input A\nII input\ninput line 3\n"
        )
        shadow_log_is_as_expected(the_big_condor, 2, 2, 6)
        lock_dir_is_clean(the_big_lock_dir)
        starter_log_is_as_expected(the_big_condor, 2, 6)


    def test_multi_cif_jobs(self, the_multi_lock_dir, the_multi_condor, completed_multi_jobs):
        output_is_as_expected(
            completed_multi_jobs[0],
            "input A\nII input\ninput line 3\n"
        )
        output_is_as_expected(
            completed_multi_jobs[1],
            "-- input A\n-- II input\n-- input line 3\n"
        )
        # This test doesn't include the delay necessary to make sure that
        # transferring common files takes longer than starting / scheduling
        # a the next shadow, so just ignore wait lines completely.
        shadow_log_is_as_expected(the_multi_condor, 4, 4, None)
        lock_dir_is_clean(the_multi_lock_dir)



#!/usr/bin/env python3
"""
Extract the docker image tagged 'ubunturoot' from nmi_tools/nmi-build-platforms.
Outputs the image name (without docker:// prefix) for use in GitHub Actions.
"""

import sys


def main():
    csv_path = 'nmi_tools/nmi-build-platforms'

    with open(csv_path, 'r') as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith('#'):
                continue

            parts = [p.strip() for p in line.split(',')]
            if len(parts) < 4:
                continue

            tags = parts[3].split(':')
            if 'ubunturoot' in tags:
                image = parts[0]
                if image.startswith('docker://'):
                    image = image[9:]
                print(image)
                return 0

    print("Error: No entry with 'ubunturoot' tag found", file=sys.stderr)
    return 1


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env pytest

import logging

from ornithology import (
    action,
    ClusterState,
    condor,
    format_script,
    write_file,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@action
def the_script(test_dir):
    the_script_file = test_dir / "the_script.py"
    contents = format_script(
    """
        #!/usr/bin/python3

        import sys
        from pathlib import Path

        Path("./failure-file-one").touch()
        Path("./failure-file-two").touch()

        sys.exit(1)
    """
    )
    write_file(the_script_file, contents)
    return the_script_file


@action
def empty_test_dir(test_dir):
        failure_file_one = test_dir / "failure-file-one"
        assert(not failure_file_one.exists())

        failure_file_two = test_dir / "failure-file-two"
        assert(not failure_file_two.exists())


@action
def the_failure_job(default_condor, the_script, test_dir):
    the_failure_job = default_condor.submit(
        {
            "executable":               the_script.as_posix(),
            "log":                      (test_dir / "job.log").as_posix(),
            "should_transfer_files":    "YES",
            "when_to_transfer_output":  "ON_SUCCESS",
            "+FailureFiles":            '"failure-file-one, failure-file-two"',
            "success_exit_code":        "0",
            "max_retries":              "0",
            "transfer_output_files":    "success-file-one",
        }
    )

    return the_failure_job


@action
def the_completed_job(the_failure_job):
    assert(the_failure_job.wait(
        condition=ClusterState.all_terminal,
        timeout=60
    ))
    return the_failure_job


class TestFailureFiles:

    def test_failure_files(empty_test_dir, the_completed_job, test_dir):
        failure_file_one = test_dir / "failure-file-one"
        assert(failure_file_one.exists())

        failure_file_two = test_dir / "failure-file-two"
        assert(failure_file_two.exists())



import sys
import os
import os.path
import logging
import subprocess
import shlex
import shutil
import tempfile
import time
import getpass


from datetime import datetime
from pathlib import Path

import htcondor2 as htcondor
import classad2 as classad

from htcondor_cli.noun import Noun
from htcondor_cli.verb import Verb
from htcondor_cli import TMP_DIR
from htcondor_cli.utils import readable_time, readable_size, s


class Create(Verb):
    """
    Creates an OCU when given an OCU request file.
    OCU request file should contain an Owner = and Request_* lines.
    """

    options = {
        "request_file": {
            "args": ("request_file",),
            "help": "Submit file",
        },
    }

    def __init__(self, logger, request_file, **options):
        # Make sure the specified request file exists and is readable
        request_file = Path(request_file)
        if not request_file.exists():
            raise FileNotFoundError(f"Could not find file: {str(request_file)}")
        if os.access(request_file, os.R_OK) is False:
            raise PermissionError(f"Could not access file: {str(request_file)}")

        # Get schedd
        schedd = htcondor.Schedd()

        # place ocu on the local schedd
        with request_file.open() as f:
            ocu_ad = classad.parseOne(f)

        try:
            result_ad = schedd.create_ocu(ocu_ad)
            if result_ad["Result"] != 0:
                raise RuntimeError(f"Failed to create ocu: {result_ad['ErrorString']}")
            ocu_id = result_ad["OCUId"]
            name = result_ad["OCUName"]
            print(f"OCU created with Id {ocu_id} and name {name}.")
        except Exception as e:
            raise RuntimeError(f"Error creating ocu: {str(e)}")

class Query(Verb):
    """
    Queries the existing OCUs in the system.
    """
    options = {
            "raw": {
                "args": ("--raw",),
                "action": "store_true",
                "default": False,
                "help": "display raw classad",
                },
            }

    def __init__(self, logger, **options):
        # Get schedd
        schedd = htcondor.Schedd()

        try:
            ad = classad.ClassAd()
            results = schedd.query_ocu(ad)

            if options["raw"]:
                for ad in results:
                    print(ad)
            else:
                print("OCUId  OCUName              Owner                    State CPUs  Memory Disk Activations")
                for ad in results:
                    ocuID = ad.get("OCUId", "unknown")
                    ocuName = ad.get("OCUName", "unknown")
                    owner = ad.get("Owner", "unknown")
                    state = ad.get("OCUState", "?")
                    cpus = ad.get("RequestCpus", 0)
                    memory = ad.get("RequestMemory", 0)
                    disk = ad.get("RequestDisk", 0)
                    activations = ad.get("OCUOwnerActivations", 0)
                    print(f"{ocuID:<6} {ocuName:<20} {owner:<24} {state:<5} {cpus:<4} {memory:<6} {disk:<4} {activations}")

        except Exception as e:
            raise RuntimeError(f"Error querying ocu: {str(e)}")

class Status(Verb):
    """
    Shows current status of OCUs
    """

    def __init__(self, logger):
        collector = htcondor.Collector()

        submitter_ads = collector.query(htcondor.AdType.Submitter, constraint="True")

        if len(submitter_ads) == 0:
            print("There are no submitters in this pool, therefore no OCUs.")
            sys.exit(0)

        print("Owner                          OCUClaimed  OCUClaimedBorrowed OCUsWantedJobs  OCURunningJobs")
        print("-----                          ----------  ------------------ --------------  --------------")

        for ad in submitter_ads:
            name = ad.get("Name", "unknown")
            ocu_requested = ad.get("OCUWantedJobs", 0)
            ocu_running = ad.get("OCURunningJobs", 0)
            ocu_claimed = ad.get("OCUClaimsClaimed", 0)
            ocu_borrowed = ad.get("OCUClaimsBorrowed", 0)

            print(f"{name:<30} {ocu_claimed:>10} {ocu_borrowed:>19} {ocu_requested:>14} {ocu_running:>14}")

class Remove(Verb):
    """
    Remove an existing OCU
    """

    options = {
        "ocu_id": {
            "args": ("ocu_id",),
            "help": "OCU ID to remove",
        },
    }

    def __init__(self, logger, ocu_id, **options):

        schedd = htcondor.Schedd()

        try:
            remove_ad = schedd.remove_ocu(int(ocu_id))
            if remove_ad["Result"] != 0:
                raise RuntimeError(f"Failed to remove ocu {ocu_id}: {remove_ad['ErrorString']}")
        except Exception as e:
            logger.error(f"Error while trying to remove ocu {ocu_id}:\n{str(e)}")
            raise RuntimeError(f"Error removing ocu: {str(e)}")

        logger.info(f"Removed {ocu_id}.")

class OCU(Noun):
    """
    Run operations on HTCondor Owner Capacity Units (OCU)
    """

    class create(Create):
        pass

    class query(Query):
        pass

    class status(Status):
        pass

    class remove(Remove):
        pass

    @classmethod
    def verbs(cls):
        return [cls.create, cls.query, cls.status, cls.remove]



import sys
import enum


class CommonFilesEventType(enum.IntEnum):
    """
    An enumeration of file transfer event types.

    .. attribute:: TransferQueued
    .. attribute:: TransferStarted
    .. attribute:: TransferFinished
    .. attribute:: WaitStarted
    .. attribute:: WaitFinished
    """

    TransferQueued   = 1
    TransferStarted  = 2
    TransferFinished = 3
    WaitStarted      = 4
    WaitFinished     = 5



import os
import re
import sys

from docutils import nodes, utils
from docutils.parsers.rst import Directive
from sphinx import addnodes
from sphinx.errors import SphinxError
from sphinx.util.nodes import split_explicit_title, process_index_entry, set_role_source_info
from htc_helpers import make_headerlink_node, make_inline_literal_node, warn

ATTRIBUTE_DEFS = {}

def classad_attribute_def_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    global ATTRIBUTE_DEFS
    app = inliner.document.settings.env.app
    docname = inliner.document.settings.env.docname

    # Create the attribute title with the classad attribute name, also containing an anchor
    anchor_title_node = make_inline_literal_node(text)

    # Create a headerlink node, which can be used to link to the anchor
    headerlink_node = make_headerlink_node(str(text), options)

    # Determine the classad type (job, submitted, collector, etc.) by ripping it out of the document title
    attr_type = ""
    type_matches = re.findall(r"/([-\w]*)-classad-attributes", docname)
    for match in type_matches:
        attr_type = match.capitalize() + " "

    if text in ATTRIBUTE_DEFS.get(attr_type, []):
        warn(f"{docname}:{lineno} | {attr_type} ClassAd attribute '{text}' already defined")
        textnode = nodes.Text(text, " ")
        return [textnode], []
    elif attr_type in ATTRIBUTE_DEFS:
        ATTRIBUTE_DEFS[attr_type].append(text)
    else:
        ATTRIBUTE_DEFS.update({attr_type : [text]})

    # Automatically include an index entry for this attribute
    index_node = addnodes.index()
    index_node['entries'] = process_index_entry(f"pair: {text} ; {attr_type} ClassAd Attribute", text)
    set_role_source_info(inliner, lineno, index_node)

    return [index_node, anchor_title_node, headerlink_node], []


def setup(app):
    app.add_role('classad-attribute-def', classad_attribute_def_role)




# Copyright 2020 HTCondor Team, Computer Sciences Department,
# University of Wisconsin-Madison, WI.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
from pathlib import Path

SCRIPT_EXT = ".cmd" if sys.platform == "win32" else ".py"

SCRIPTS = {
    path.stem: path.with_suffix(SCRIPT_EXT).as_posix()
    for path in Path(__file__).parent.iterdir()
    if path.stem != "__init__"
}

WIN32_PY_SCRIPT_PROLOG = [
    "0<0# : ^", "'''", "@echo off",
    ("py.exe" if not sys.executable else '"{}"'.format(sys.executable)) + ' "%~f0" %*',
    "@goto :EOF", "'''"
]

def prepare_script(path):
    if os.path.isfile(path) : return

    source = Path(path).with_suffix(".py")
    if os.path.isfile(source):
        try:
            with open(source, 'rb') as f:
                body = f.read()
            if body:
                with open(path,'wb') as f:
                    f.write("\n".join(WIN32_PY_SCRIPT_PROLOG).encode('utf8') + body)
                os.remove(source)
        except IOError:
            return # do nothing
    return


# Return a space separated list of all custom
# fto plugins to be readily available to all tests
def custom_fto_plugins() -> str:
    if SCRIPT_EXT == ".cmd":
       for name,path in SCRIPTS.items():
          prepare_script(path)

    return " ".join([
        SCRIPTS["null_plugin"],  # null://
        SCRIPTS["debug_plugin"],  # debug://, encode://
    ])



#!/usr/bin/env pytest

import pytest
from ornithology import *
import os
import json
from shutil import rmtree

#------------------------------------------------------------------
# Magical class to cd into specific dir and return to starting dir
# via pythons 'with' keyword
class ChangeDir:
    def __init__(self, switch):
        self.orig = None
        self.switch = switch

    def __enter__(self):
        self.orig = os.getcwd()
        os.chdir(self.switch)

    def __exit__(self, exec_type, exec_value, exec_traceback):
        os.chdir(self.orig)
        self.orig = None

#------------------------------------------------------------------
def setup_use_dag_dir():
    SUBDIR = "subdir"
    os.mkdir(SUBDIR)

    with ChangeDir(SUBDIR):
        with open("test.dag", "w") as f:
            f.write("INCLUDE info.inc\n")

        with open("info.inc", "w") as f:
            f.write("JOB A desc\n")

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def setup_complex_dag():
    with open("sub-splice.dag", "w") as f:
        f.write("JOB A desc\n")

    with open("splice.dag", "w") as f:
        f.write("""# Splice DAG
JOB A desc
JOB B desc

SPLICE DEEP sub-splice.dag

PARENT A CHILD DEEP
""")

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def setup_validation_errors():
    with open("single-node.dag", "w") as f:
        f.write("JOB A desc\n")

    with open("recursive.dag", "w") as f:
        f.write("SPLICE UH-OH recursive.dag\n")

    with open("recursive.inc", "w") as f:
        f.write("INCLUDE recursive.inc\n")

    with open("splice-with-final.dag", "w") as f:
        f.write("FINAL F desc\n")

    SUBDIR = "subdir"

    os.mkdir(SUBDIR)

    with open("splice.dag", "w") as f:
        f.write(f"SPLICE S sub-splice.dag DIR {SUBDIR}\n")

    with ChangeDir(SUBDIR):
        with open("sub-splice.dag", "w") as f:
            f.write("SPLICE S sub-sub-splice.dag\n")

        with open("sub-sub-splice.dag", "w") as f:
            f.write("JOB\n")

#------------------------------------------------------------------
KEY_DAG = "DagFile"
KEY_SETUP = "SetupFunc"
KEY_EXIT = "ExitCode"
KEY_ERRORS = "Errors"
KEY_STATS = "Stats"
KEY_OPTIONS = "ExtraCmdOptions"
KEY_OVERRIDE = "OverrideRootDagFile"

TEST_CASES = {
    # Test Case: Check parser syntax errors (note full parser syntax errors checked in unit test)
    "PARSE_ERRORS": {
        KEY_DAG: """# DAG riddled with syntax errors
JOB all_nodes desc
FINAL FOO+BAR desc
RETRY
DNE blah blah blah
""",
        KEY_EXIT: 1,
        KEY_ERRORS: [
            {
                "Reason": "Node name is a reserved word",
                "SourceFile": "PARSE_ERRORS.dag",
                "SourceLine": 2,
                "DagCommand": "JOB",
            },
            {
                "Reason": "Node name contains illegal charater",
                "SourceFile": "PARSE_ERRORS.dag",
                "SourceLine": 3,
                "DagCommand": "FINAL",
            },
            {
                "Reason": "No node name specified",
                "SourceFile": "PARSE_ERRORS.dag",
                "SourceLine": 4,
                "DagCommand": "RETRY",
            },
            {
                "Reason": "'DNE' is not a valid DAG command",
                "SourceFile": "PARSE_ERRORS.dag",
                "SourceLine": 5,
            },
        ],
    },
    # Test Case: Check post syntax parse validation errors
    "VALIDATION_ERRORS": {
        KEY_DAG: """# Syntactically correct DAG with validation issues
FINAL F desc
FINAL F2 desc
PROVISIONER P desc
PROVISIONER P2 desc
SERVICE S desc

VARS ALL_NODES foo="bar"
DONE ALL_NODES

SPLICE NO_FILE dne.dag
SPLICE S1 single-node.dag
SPLICE S2 splice-with-final.dag
SPLICE S3 splice.dag
SPLICE RECURSIVE recursive.dag

INCLUDE dne.inc
INCLUDE recursive.inc

JOB A desc
JOB A desc
JOB S1 desc
SPLICE A single-node.dag

PARENT DNE CHILD A
PARENT A CHILD DNE
PARENT DNE CHILD DNE
PARENT F P S CHILD A

RETRY S1 3
PRIORITY DNE 5

CATEGORY DNE FOO

CONFIG /dne/first.conf
CONFIG second.conf

REJECT
""",
        KEY_SETUP: setup_validation_errors,
        KEY_EXIT: 1,
        KEY_ERRORS: [
            {
                "Reason": "Final node is already defined in the DAG",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 3,
                "DagCommand": "FINAL",
            },
            {
                "Reason": "Provisioner node is already defined in the DAG",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 5,
                "DagCommand": "PROVISIONER",
            },
            {
                "Reason": "ALL_NODES can not be used with DONE command",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 9,
                "DagCommand": "DONE",
            },
            {
                "Reason": "Failed to read splice NO_FILE: Provided file path does not exist",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 11,
                "DagCommand": "SPLICE",
            },
            {
                "Reason": "Spliced DAGs can not contain a final node",
                "SourceFile": "splice-with-final.dag",
                "SourceLine": 1,
                "DagCommand": "FINAL",
            },
            {
                "Reason": "Missing node name",
                "SourceFile": "sub-sub-splice.dag",
                "SourceLine": 1,
                "DagCommand": "JOB"
            },
            {
                "Reason": "Recursive splicing detected",
                "SourceFile": "recursive.dag",
                "SourceLine": 1,
                "DagCommand": "SPLICE",
            },
            {
                "Reason": "Failed to read include file: Provided file path does not exist",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 17,
                "DagCommand": "INCLUDE",
            },
            {
                "Reason": "Recursive file inclusion detected",
                "SourceFile": "recursive.inc",
                "SourceLine": 1,
                "DagCommand": "INCLUDE",
            },
            {
                "Reason": "Node A is already defined in the DAG",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 21,
                "DagCommand": "JOB",
            },
            {
                "Reason": "Node S1 is already defined in the DAG",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 22,
                "DagCommand": "JOB",
            },
            {
                "Reason": "Splice name A is already defined in the DAG",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 23,
                "DagCommand": "SPLICE",
            },
            {
                "Reason": "References to undefined nodes: Parents [DNE]",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 25,
                "DagCommand": "PARENT",
            },
            {
                "Reason": "References to undefined nodes: Children [DNE]",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 26,
                "DagCommand": "PARENT",
            },
            {
                "Reason": "References to undefined nodes: Parents [DNE] Children [DNE]",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 27,
                "DagCommand": "PARENT",
            },
            {
                "Reason": "FINAL node F can not have dependencies",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 28,
                "DagCommand": "PARENT",
            },
            {
                "Reason": "PROVISIONER node P can not have dependencies",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 28,
                "DagCommand": "PARENT",
            },
            {
                "Reason": "SERVICE node S can not have dependencies",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 28,
                "DagCommand": "PARENT",
            },
            {
                "Reason": "Cannot be applied to splice S1",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 30,
                "DagCommand": "RETRY",
            },
            {
                "Reason": "References undefined node DNE",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 31,
                "DagCommand": "PRIORITY",
            },
            {
                "Reason": "References to undefined nodes: DNE",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 33,
                "DagCommand": "CATEGORY",
            },
            {
                "Reason": "Configuration file /dne/first.conf does not exist",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 35,
                "DagCommand": "CONFIG",
            },
            {
                "Reason": "DAG configuration file is already defined",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 36,
                "DagCommand": "CONFIG",
            },
            {
                "Reason": "DAG marked with REJECT command",
                "SourceFile": "VALIDATION_ERRORS.dag",
                "SourceLine": 38,
            },
        ],
    },
    # Test Case: Provided DAG file DNE
    "NO_DAG": {
        KEY_EXIT: 1,
        KEY_ERRORS: [
            {
                "Reason": "Failed to open DAG NO_DAG.dag: Provided file path does not exist",
                "SourceFile": "NO_DAG.dag",
                "SourceLine": 0,
            }
        ]
    },
    # Test Case: Statistics for shiskabob DAG
    "SHISKABOB": {
        KEY_DAG: """#Shiskabob DAG
JOB A desc
JOB B desc
JOB C desc

PARENT A CHILD B
PARENT B CHILD C
""",
        KEY_EXIT: 0,
        KEY_STATS: {
            "TotalNodes": 3,
            "NumNodes": 3,
            "NumArcs": 2,
            "GraphHeight": 3,
            "GraphWidth": 1,
            "WidthByDepth": [1,1,1],
        }
    },
    # Test Case: Statistics for diamond DAG
    "DIAMOND": {
        KEY_DAG: """# Classic diamond DAG shape
JOB A desc
JOB B desc
JOB C desc
JOB D desc

PARENT A CHILD B C
PARENT B C CHILD D
""",
        KEY_EXIT: 0,
        KEY_STATS: {
            "TotalNodes": 4,
            "NumNodes": 4,
            "NumArcs": 4,
            "GraphHeight": 3,
            "GraphWidth": 2,
            "WidthByDepth": [1,2,1],
        }
    },
    # Test Case: Statistics for 'complex' DAG
    "COMPLEX_STATS": {
        KEY_DAG: """# Complex DAG Structure
FINAL fin desc
PROVISIONER BANK desc

SERVICE SER-1 desc
SERVICE SER-2 desc
SERVICE SER-3 desc

JOB A desc
JOB B desc
JOB C desc
JOB D desc

SUBDAG EXTERNAL SUB-1 sub.dag
SUBDAG EXTERNAL SUB-2 sub.dag
SUBDAG EXTERNAL SUB-3 sub.dag

SPLICE S1 splice.dag
SPLICE S2 splice.dag

PARENT A CHILD B
PARENT B CHILD C D
PARENT C D CHILD SUB-1 SUB-2 SUB-3
PARENT SUB-1 SUB-2 SUB-3 CHILD S1 S2
""",
        KEY_SETUP: setup_complex_dag,
        KEY_EXIT: 0,
        KEY_STATS: {
            "TotalNodes": 20,
            "NumNodes": 10,
            "NumJoinNodes": 2,
            "NumServiceNodes": 3,
            "NumSubDAGs": 3,
            "NumSplices": 4,
            "FinalNode": "fin",
            "ProvisionerNode": "BANK",
            "NumArcs": 17,
            "GraphHeight": 8,
            "GraphWidth": 4,
            "WidthByDepth": [1,1,2,1,3,1,4,2],
        }
    },
    # Test Case: Check for cycle in DAG (No Initial nodes)
    "CYCLE_NO_INITIAL": {
        KEY_DAG: """# DAG cycle with no initial nodes
JOB A desc
JOB B desc

PARENT A CHILD B
PARENT B CHILD A
""",
        KEY_EXIT: 1,
        KEY_STATS: {
            "TotalNodes": 2,
            "NumNodes": 2,
            "NumArcs": 2,
            "IsCyclic": True,
            "DetectedCycle": "No initial nodes discovered",
        },
    },
    # Test Case: Check for cycle in DAG (Disjointed section has no begining)
    "CYCLE_DISJOINTED": {
        KEY_DAG: """# DAG with disjointed cyclic portion
JOB A desc
JOB B desc
JOB C desc

PARENT A CHILD B
PARENT B CHILD A
""",
        KEY_EXIT: 1,
        KEY_STATS: {
            "TotalNodes": 3,
            "NumNodes": 3,
            "NumArcs": 2,
            "IsCyclic": True,
            "DetectedCycle": "Failed to visit nodes A,B",
        }
    },
    # Test Case: Check for cycle in DAG (Known dependency chain that loops)
    "CYCLE_CHAIN": {
        KEY_DAG: """# Cyclic DAG with well defined chain
JOB A desc
JOB B desc
JOB C desc

PARENT A CHILD B
PARENT B CHILD C
PARENT C CHILD B
""",
        KEY_EXIT: 1,
        KEY_STATS: {
            "TotalNodes": 3,
            "NumNodes": 3,
            "NumArcs": 3,
            "IsCyclic": True,
            "DetectedCycle": "A -> B -> C -> B",
        }
    },
    # Test Case: Verify join nodes
    "OPT_USE_JOIN_STATS": {
        KEY_DAG: """# Simple DAG that makes a join node
JOB A desc
JOB B desc
JOB C desc
JOB D desc
JOB E desc
JOB F desc

PARENT A B C CHILD D E F
""",
        KEY_OPTIONS: ["-joinNodes"],
        KEY_EXIT: 0,
        KEY_STATS: {
            "TotalNodes": 7,
            "NumNodes": 6,
            "NumJoinNodes": 1,
            "NumArcs": 6,
            "GraphHeight": 3,
            "GraphWidth": 3,
            "WidthByDepth": [3,1,3],
        }
    },
    # Test Case: Verify stats for previous DAG with no join nodes
    "OPT_NO_JOIN_STATS": {
        KEY_DAG: """# Simple DAG that makes a join node
JOB A desc
JOB B desc
JOB C desc
JOB D desc
JOB E desc
JOB F desc

PARENT A B C CHILD D E F
""",
        KEY_OPTIONS: ["-NoJoinNodes"],
        KEY_EXIT: 0,
        KEY_STATS: {
            "TotalNodes": 6,
            "NumNodes": 6,
            "NumArcs": 9,
            "GraphHeight": 2,
            "GraphWidth": 3,
            "WidthByDepth": [3,3],
        }
    },
    # Test Case: Verify command option for allow illegal characters
    "OPT_ALLOW_ILLEGAL_CHARS": {
        KEY_DAG: """# Node with illegal character(s)
JOB A+B desc
""",
        KEY_OPTIONS: ["-AllowIllegalChars"],
        KEY_EXIT: 0,
    },
    # Test Case: Verify failure when specifying multiple printers
    "OPT_ERROR_MULTIPLE_PRINTERS": {
        KEY_OPTIONS: ["-stat"],
        KEY_EXIT: 2,
    },
    # Test Case: Verify command option for using DAG directory
    "OPT_USE_DAG_DIR": {
        KEY_EXIT: 0,
        KEY_SETUP: setup_use_dag_dir,
        KEY_OPTIONS: ["-UseDagDir"],
        KEY_OVERRIDE: "subdir/test.dag",
        KEY_STATS: {
            "TotalNodes": 1,
            "NumNodes": 1,
            "GraphHeight": 1,
            "GraphWidth": 1,
            "WidthByDepth": [1],
        }
    },
}

#------------------------------------------------------------------
@action(params={name: name for name in TEST_CASES})
def run_command(default_condor, test_dir, request):
    CASE = request.param

    # make test case directory
    case_dir = test_dir / CASE
    if case_dir.exists():
        rmtree(case_dir)
    os.mkdir(str(case_dir))

    # Get test case details
    DETAILS = TEST_CASES[CASE]

    # Change into test case directory
    with ChangeDir(CASE):
        # Root test case dag file <CASE>.dag
        DAG = DETAILS.get(KEY_OVERRIDE, f"{CASE}.dag")

        # Write DAG file contents (unless specified as Nonde)
        if DETAILS.get(KEY_DAG) is not None:
            with open(DAG, "w") as f:
                f.write(DETAILS[KEY_DAG])

        # Run specific test case setup function
        setup = DETAILS.get(KEY_SETUP)
        if setup is not None:
            setup()

        # Setup command to execute
        CMD = ["condor_dag_checker", "-json", DAG]
        CMD.extend(DETAILS.get(KEY_OPTIONS, []))

        # Run command
        p = default_condor.run_command(CMD)

        # Dump command stdout for debugging
        with open("cmd.out", "w") as f:
            f.write(p.stdout)

        # Dump command stderr for debugging
        with open("cmd.err", "w") as f:
            f.write(p.stderr)

    # Pass all test case details needed for checks
    return (DAG, p, DETAILS[KEY_EXIT], DETAILS.get(KEY_ERRORS), DETAILS.get(KEY_STATS))

#------------------------------------------------------------------
@action
def test_dag(run_command):
    return run_command[0]

@action
def process(run_command):
    return run_command[1]

@action
def expected_exit_code(run_command):
    return run_command[2]

@action
def expected_errors(run_command):
    return run_command[3]

@action
def expected_stats(run_command):
    return run_command[4]

#==================================================================
class TestDagCheckerTool:
    def test_check_exit_code(self, process, expected_exit_code):
        """Verify command produced proper exit code"""
        assert process.returncode == expected_exit_code

    def test_check_errors(self, process, expected_errors):
        """Check for all expected error messages"""
        has_json = True
        try:
            RESULT = json.loads(process.stdout)[0]
        except json.decoder.JSONDecodeError:
            has_json = False

        if expected_errors is not None:
            assert len(expected_errors) == RESULT.get("NumErrors")

            for error in RESULT["Errors"]:
                idx = expected_errors.index(error)
                del expected_errors[idx]

            assert len(expected_errors) == 0
        elif has_json:
            assert RESULT.get("NumErrors") == None
            assert RESULT.get("Errors") == None

    def test_check_stats(self, test_dag, process, expected_stats):
        """Check for specific statistics"""
        has_json = True
        try:
            RESULT = json.loads(process.stdout)[0]
        except json.decoder.JSONDecodeError:
            has_json = False

        if expected_stats is not None:
            for STAT, VAL in expected_stats.items():
                assert STAT in RESULT
                assert expected_stats[STAT] == RESULT[STAT]
        elif has_json:
            assert KEY_DAG in RESULT
            assert RESULT[KEY_DAG] == test_dag




#!/usr/bin/env pytest

import logging

import htcondor2 as htcondor

from ornithology import *

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

"""
This test creates a new X.509 user credential, submits a job that uses that
credential, and verfies that HTCondor can delegate the credential.
"""

@action
def path_to_the_x509_cred(default_condor, test_dir):
    # If openssl is told to write both private key and public cert to the
    # same file, it'll put the key first.
    # The standard proxy file layout puts the certificate first and our code
    # to read a proxy file will fail if that ordering isn't maintained.
    # Thus, we need to manually assemble the file here to have the cert
    # appear first.
    key_path = test_dir / "test.key"
    cred_path = test_dir / "test.pem"
    rv = default_condor.run_command(
        ["openssl", "req", "-nodes", "-x509", "-new", "-newkey", "rsa:4096", "-keyout", key_path.as_posix(), "-out", cred_path.as_posix(), "-days", "365", "-subj", "/C=US/ST=Wisconsin/L=Madison/O=HTCondor/CN=Test Credential"],
        timeout=90,
    )
    assert rv.returncode == 0
    key_data = key_path.read_text()
    with open(cred_path, 'a') as cred_file:
        cred_file.write(key_data)
    cred_path.chmod(0o600)
    return cred_path

@action
def path_to_the_job_script(test_dir):
    path = test_dir / "job.sh"
    script = """
        #!/bin/bash
        if [ -z "$X509_USER_PROXY" ] ; then
            echo "X509_USER_PROXY not defined"
            exit 1
        fi
        if [ ! -r "$X509_USER_PROXY" ] ; then
            echo "X509_USER_PROXY isn't a readable file"
            exit 2
        fi
        proxy_subj=`openssl x509 -in $X509_USER_PROXY -noout -subject -nameopt rfc2253`
        if [ $? != 0 ] ; then
            echo "openssl failed"
            exit 3
        fi
        if ! echo "$proxy_subj" | grep -q 'subject= *CN=[0-9]*,CN=Test' ; then
            echo "Proxy subject doesn't look valid:"
            echo $proxy_subj
            exit 4
        fi
        exit 0
        """
    write_file(path, format_script(script))
    return path

class TestX509Delegation:
    def test_delegate_job_x509_credential(self, default_condor, path_to_the_x509_cred, path_to_the_job_script):
        job = default_condor.submit(
            {
                "universe": "vanilla",
                "executable": path_to_the_job_script.as_posix(),
                "output": "job.out",
                "error": "job.err",
                "log": "job.log",
                "should_transfer_files": "YES",
                "when_to_transfer_output": "ON_EXIT",
                "x509userproxy": path_to_the_x509_cred.as_posix(),
                "leave_in_queue": "true",
            }
        )
        assert job.wait(condition=ClusterState.all_complete, timeout=60)

        assert job.state[0] == JobStatus.COMPLETED

        job_ad = job.query()[0]
        exit_code = job_ad['ExitCode']
        assert exit_code == 0



#!/usr/bin/env pytest

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

from ornithology import (
    action,
)

from htcondor2._schedd import (
    job_spec_hack
)

from classad2 import (
    ExprTree,
)


def f_success(* pargs):
    assert(True)


def f_failure(* pargs):
    assert(False)


class TestJobSpecTypes:

    def test_job_spec_int(self):
        job_spec_hack( "addr", 7, f_failure, f_success, [] )


    def test_job_spec_str(self):
        job_spec_hack( "addr", "6", f_failure, f_success, [] )
        job_spec_hack( "addr", "6.5", f_success, f_failure, [] )
        job_spec_hack( "addr", "ClusterID >= 4", f_failure, f_success, [] )


    def test_job_spec_ExprTree(self):
        constraint = ExprTree("ClusterID != 3")
        job_spec_hack( "addr", constraint, f_failure, f_success, [] )


    def test_job_spec_list(self):
        job_spec_hack( "addr", ["2", "1.1", "10"], f_failure, f_success, [] )


    # There's probably a way to mark this as expected to fail with a
    # particular exception, but for a test this simple, I won't bother.
    def test_job_spec_true_negative(self):
        try:
            job_spec_hack( "addr", 3.141, f_failure, f_failure, [] )
            assert(False)
        except TypeError:
            pass



from flask import Blueprint, render_template, request, current_app
import csv
import urllib.request
import functools
import xml.etree.ElementTree as ET
from utils import cache_response_to_disk, make_data_response, getOrganizationFromInstitutionID
import time
from blueprints.landing import get_ce_facility_site_descrip, get_next_prev_sites

#######################
# Functions to query Topology and cache responses
#######################

# Query the Topology database for project info, storing the result in a local file for 1 hour
# We store it in a local file so that we can share the response between multiple wsgi server
# processes, and to avoid hitting the topology for each process.
@cache_response_to_disk(file_name="topology_projects.bin", seconds_to_cache=60*60)
def _get_topology_projects_xml():
    # may raise urllib.error.HTTPError

    handle = urllib.request.urlopen("https://topology.opensciencegrid.org/miscproject/xml")
    result_bytes = handle.read()
    return result_bytes

# Parse the XML response from the Topology database and cache the result
# for 1 hour. How does the caching for one hour work? The function is
# decorated with @functools.lru_cache(maxsize=1), which caches the result
# of the function call. The cache is invalidated after one hour by
# using the hour_counter parameter. The function is called with the current
# hour as the parameter, and the result is cached for that hour. When the
# hour changes, the cache is invalidated and a new result is generated.
@functools.lru_cache(maxsize=1)
def _get_projects_tree(hour_counter: int) -> ET.Element:
    hour_counter += 0  # just to avoid warning about unused variable
    result_bytes, _ = _get_topology_projects_xml()
    # Decode the bytes to a string using surrogateescape error handling
    result_string = result_bytes.decode(
        encoding="utf-8", errors="surrogateescape"
    )
    # Parse the XML response and return the root element
    return ET.fromstring(result_string)

# Cache the hash for one hours using the same hour_counter trick as above.
# Cache is maxsize 3, so that we can cache for field = "FieldOfScience"
# and field = "Organization" and field = "InstitutionID".
# The cache is invalidated after one hour by using the hour_counter parameter.
# The function is called with the current hour as the parameter, and the result
# is cached for that hour. When the hour changes, the cache is invalidated and a new result is generated.
@functools.lru_cache(maxsize=3)
def _get_project_fos_hash(hour_counter: int, field: str):
    def safe_elem_text(elem: ET.Element) -> str:
        try:
            return elem.text.strip()
        except AttributeError:  # we got a None
            return ""

    project_hash = {}

    projects_tree = _get_projects_tree(hour_counter)
    for project in projects_tree.findall("./Project"):
        name = safe_elem_text(project.find("./Name"))
        if not name:
            continue  # malformed project?
        field_of_science = safe_elem_text(project.find("./" + field))
        project_hash[name] = field_of_science

    return project_hash


def organization_from_project(project_name: str):
    hour_counter = int(time.time() / 3600)
    organizationName =  _get_project_fos_hash(hour_counter,"Organization").get(project_name, "Unknown").replace(',',' -')
    institutionId =  _get_project_fos_hash(hour_counter,"InstitutionID").get(project_name, "Unknown").replace(',',' -')
    # We want to get the organization name from the institution ID, as this is the 'new' way to do it.
    # However, if there is not a valid institution ID, we will fallback to the organization name.
    return getOrganizationFromInstitutionID(institutionId,organizationName)


def field_of_science_from_project(project_name: str):
    hour_counter = int(time.time() / 3600)
    return _get_project_fos_hash(hour_counter,"FieldOfScience").get(project_name, "Unknown").replace(',',' -')


#######################
# Functions to query Ganglia Web and format response for HTCondorView
#######################

@cache_response_to_disk()
def get_data_from_ganglia():
    # Grab parameters host (CE name), r (range: day, week, or month)
    host = request.args.get('host','chtc-spark-ce1.svc.opensciencegrid.org')
    r = request.args.get('r','day')

    # Grab data as a raw csv from ganglia, which was stashed there by the condor_gangliad
    import pandas as pd
    # If host is not fully qualified, add the default domain
    if not '.' in host:
        host = host + '.' + current_app.config['CE_DASHBOARD_DEFAULT_CE_DOMAIN']
    
    metricsd_url = current_app.config['CE_DASHBOARD_METRICSD_URL']

    # Retrieve any metrics from Ganglia that start with any of the following phrases
    metric_exprs = ['Cpus', 'Gpus','Memory','Disk','Bcus','EPs']
    metric_args = '&'.join([f'mreg[]=%5E{expr}' for expr in metric_exprs])

    df=pd.read_csv(f'{metricsd_url}/ganglia/graph.php?r={r}&hreg[]={host}&{metric_args}&aggregate=1&csv=1',skipfooter=1,engine='python')

    # Transpose the data received from ganglia into the format we need for the CE Dashboard frontend

    # Rename 'Timestamp' column to 'Date' for clarity and set it as the index
    df.rename({'Timestamp':'Date'}, axis='columns', inplace=True)
    df.set_index('Date', inplace=True)
    # Split column names into multiple levels based on specific delimiters
    df.columns = df.columns.str.split(' |_Project |Use ', n=2, expand=True)
    # Define new names for the column levels
    new_names = ['Host', 'Resource', 'Project']
    df.columns.names = new_names
    # Reshape the DataFrame from wide to long format
    df = df.stack(new_names).rename('value').reset_index()
    # Drop the 'Host' column as it is no longer needed
    df.drop(columns=['Host'], inplace=True)
    # Strip whitespace from 'Project' column and handle empty project names
    df['Project'] = df['Project'].map(str.strip)
    df.loc[df['Project'] == '', 'Project'] = "____meta_" + df['Resource']
    # Simplify 'Resource' column by removing specific suffixes
    df['Resource'] = df['Resource'].str.replace('(In|NotIn)$', '', regex=True)
    # Pivot the DataFrame to create a schema with 'Date' and 'Project' as indices
    df = df.pivot(index=['Date', 'Project'], columns='Resource', values='value')
    # Fill missing values with zeros to handle incomplete data
    df.fillna(0, inplace=True)
    # Remove axis names and reset the index for a flat structure
    df = df.rename_axis(None, axis=1).reset_index()
    # Group by 'Date' and 'Project' to aggregate data
    df = df.groupby(['Date', 'Project'], as_index=False).sum()
    # Add a 'Type' column to distinguish between project usage (with P) and total resource (with R) entries
    df['Type'] = df['Project'].apply(lambda x: 'R' if x.startswith('____meta_') else 'P')
    # Replace meta entries with more descriptive names for better readability
    for oldname, newname in (['CpusIn', 'CpusAllocated'],
                              ['CpusNotIn', 'CpusUnallocated'],
                              ['GpusIn', 'GpusAllocated'],
                              ['GpusNotIn', 'GpusUnallocated'],
                              ['MemoryIn', 'MemoryAllocated'],
                              ['MemoryNotIn', 'MemoryUnallocated'],
                              ['DiskIn', 'DiskAllocated'],
                              ['DiskNotIn', 'DiskUnallocated'],
                              ['BcusIn', 'BcusAllocated'],
                              ['BcusNotIn', 'BcusUnallocated'],
                              ['EPsRunning', 'EPsRunning'],
                              ['EPsCpuLimited', 'EPsCpuLimited'],
                              ['EPsMemoryLimited', 'EPsMemoryLimited'],
                              ['EPsDiskLimited', 'EPsDiskLimited']):
        df['Project'] = df['Project'].str.replace(f'____meta_{oldname}', newname)
    # Get rid of columns that are not needed; specifically, we don't want info per user, just per project   
    df.drop(columns=['Cpus_User'],inplace=True,errors='ignore')
    df.drop(columns=['Memory_User'],inplace=True,errors='ignore')
    # Remove projects where all rows with Type="P" have Cpus=0.0
    projects_to_remove = df[(df['Type'] == 'P') & (df['Cpus'] == 0.0)]['Project'].unique()
    projects_with_nonzero_cpus = df[(df['Type'] == 'P') & (df['Cpus'] > 0.0)]['Project'].unique()
    projects_to_remove = set(projects_to_remove) - set(projects_with_nonzero_cpus)
    df = df[~df['Project'].isin(projects_to_remove)]
    # Insert Field and Organization columns, which we can figure out from the project name via Topology
    df['Field'] = df['Project'].apply(field_of_science_from_project)
    df['Organization'] = df['Project'].apply(organization_from_project)

    # Finally, serialize the data into a string in jsonp format as expected by the HTCondorView javascript

    mycsv = df.to_csv(index=False,quoting=csv.QUOTE_ALL).splitlines()
    index = 0
    response = "["
    while index < len(mycsv) - 1:
        response += "[" + mycsv[index] + "],\n"
        index += 1
    response += "[" + mycsv[index] + "]\n]"
    
    return response

#######################
# Flask routes
#######################

overview_bp = Blueprint('overview', __name__)

overview_linkmap = {
    'Overview': 'overview.html',
    'Contributed': 'contributed.html',
}

@overview_bp.route('/allocated_graph_<resource>.html')
def handle_allocated_graph(resource):
    host = request.args.get('host','chtc-spark-ce1.svc.opensciencegrid.org')
    from blueprints.landing import get_ce_facility_site_descrip
    facility, site, *_ = get_ce_facility_site_descrip(host)
    title = facility if facility!='Unknown' else host
    return render_template('allocated_graph.html', resource=resource,
                           linkmap=overview_linkmap,
                           page_title=title, ce_facility_name=facility, ce_site_name=site)

@overview_bp.route('/contributed.html')
def handle_contributed():
    time_range = request.args.get('r','week')
    if time_range not in ['hour','day','week','month','year']:
        time_range = 'week'
    host = request.args.get('host','chtc-spark-ce1.svc.opensciencegrid.org')
    facility, site, name, *_ = get_ce_facility_site_descrip(host)
    next_site, prev_site = get_next_prev_sites(host)
    title = facility if facility!='Unknown' else host

    return render_template('contributed.html', host=host, 
                           linkmap=overview_linkmap, time_range=time_range,
                           page_title=title, page_subtitle = name, ce_facility_name=facility, ce_site_name=site,
                           next_page=next_site, prev_page=prev_site)

@overview_bp.route('/overview.html')
def overview():
    host = request.args.get('host','chtc-spark-ce1.svc.opensciencegrid.org')
    facility, site, name, descrip, health = get_ce_facility_site_descrip(host)
    next_site, prev_site = get_next_prev_sites(host)
    title = facility if facility!='Unknown' else host
    return render_template('overview.html', 
                           page_title=title, page_subtitle = name, ce_facility_name=facility, ce_site_name=site,
                           ce_description=descrip, ce_health=health, linkmap=overview_linkmap, 
                           next_page=next_site, prev_page=prev_site)

@overview_bp.route('/data/ce_overview')
def ce_overview_data():
    response_body, cached_response = get_data_from_ganglia()
    return make_data_response(response_body,cached_response)

@overview_bp.route('/error-no-data.html')
def error_no_data():
    from blueprints.landing import get_ce_facility_site_descrip
    host = request.args.get('host','chtc-spark-ce1.svc.opensciencegrid.org')
    facility, site, name, *_ = get_ce_facility_site_descrip(host)
    next_site, prev_site = get_next_prev_sites(host)
    msg = request.args.get('msg','No data available')
    return render_template('error_no_data.html', 
                           page_title=facility, page_subtitle = name, 
                           ce_name=site, errMsg=msg, linkmap=overview_linkmap,
                           next_page=next_site, prev_page=prev_site)



import sys
import enum


class SubmitMethod(enum.IntEnum):
    """
    An enumeration of reserved values for submit methods.

    .. attribute:: MinReserved

    .. attribute:: CondorSubmit

    .. attribute:: DAGMan

    .. attribute:: PythonBindings

    .. attribute:: HTCondorJobSubmit

    .. attribute:: HTCondorDagSubmit

    .. attribute:: HTCondorJobSetSubmit

    .. attribute:: UserSet

    .. attribute:: MaxReserved
    """

    CondorSubmit            = 0
    DAGMan                  = 1
    PythonBindings          = 2
    HTCondorJobSubmit       = 3
    HTCondorDagSubmit       = 4
    HTCondorJobSetSubmit    = 5
    UserSet                 = 100

    MinReserved             = 0
    MaxReserved             = 100



import os


def try_os_setegid(number=None, name=None):
    # This can happen if we're calling this function with its previous result.
    if number is None and name is None:
        return None

    try:
        from grp import getgrnam

        egid = os.getegid()
        if number is None:
            number = getgrnam(name).gr_gid
        os.setegid(number)
        return egid
    except ModuleNotFoundError:
        return None
    except PermissionError:
        return None
    except KeyError:
        return None


def try_os_seteuid(number=None, name=None):
    # This can happen if we're calling this function with is previous result.
    if number is None and name is None:
        return None


    try:
        from pwd import getpwnam

        euid=os.geteuid()
        if number is None:
            number = getpwnam(name).pw_uid
        os.seteuid(number)
        return euid
    except ModuleNotFoundError:
        return None
    except PermissionError:
        return None
    except KeyError:
        return None



#!/usr/bin/env pytest

import os
import logging
import platform
import time
import htcondor2 as htcondor

from ornithology import (
    config,
    standup,
    action,
    Condor,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

@standup
def condor(test_dir):
    with Condor(
        local_dir=test_dir / "condor",
        config={
            "COLLECTOR.ALLOW_ADMINISTRATOR": "$(ALLOW_ADMINISTRATOR) joe@here",
            "MASTER_NAME": "test@here",
            "STARTD_NAME": "test@here",
            "MASTER.SEC_READ_AUTHENTICATION": "REQUIRED",
            "MASTER.ALLOW_READ": "$(ALLOW_DAEMON)",
        },
    ) as condor:
        yield condor

@action
def token(condor):
    rv = condor.run_command(
        ["condor_token_create", "-identity", "joe@here", "-token", "joe"]
    )
    assert rv.returncode == 0
    return True

class TestAdminCapability:
    def test_reconfig(self, condor, token):
        # Ugh. The master waits for 5 seconds before sending its ad to
        # the collector, and Ornithology doesn't wait for the master ad
        # to appear before assuming the installation is up and ready for
        # use.
        time.sleep(5)

        # Override SEC_PASSWORD_DIRECTORY so that condor_reconfig doesn't
        #   have access to the POOL signing key.
        # Override SEC_CLIENT_AUTHENTICATION_METHODS so that our special
        #   IDToken is used for authentication.
        # Override MASTER_ADDRESS_FILE so that condor_config_val has to
        #   query the collector to locate the master.
        rv = condor.run_command(
            ["env", "_condor_SEC_PASSWORD_DIRECTORY=/nothere",
                "_condor_SEC_CLIENT_AUTHENTICATION_METHODS=IDTOKEN",
                "_condor_MASTER_ADDRESS_FILE=/nothere",
                "condor_reconfig", "-name", "test@here"]
        )
        assert rv.returncode == 0

    def test_drain(self, condor, token):
        # Override SEC_PASSWORD_DIRECTORY so that condor_drain doesn't
        #   have access to the POOL signing key.
        # Override SEC_CLIENT_AUTHENTICATION_METHODS so that our special
        #   IDToken is used for authentication.
        # Override STARTD_ADDRESS_FILE so that condor_drain has to
        #   query the collector to locate the startd.
        rv = condor.run_command(
            ["env", "_condor_SEC_PASSWORD_DIRECTORY=/nothere",
                "_condor_SEC_CLIENT_AUTHENTICATION_METHODS=IDTOKEN",
                "_condor_STARTD_ADDRESS_FILE=/nothere",
                "condor_drain", "-debug", "slot1@test@here"]
        )
        assert rv.returncode == 0

    def test_config_val(self, condor, token):
        # Ugh. The master waits for 5 seconds before sending its ad to
        # the collector, and Ornithology doesn't wait for the master ad
        # to appear before assuming the installation is up and ready for
        # use.
        time.sleep(5)

        # Override SEC_PASSWORD_DIRECTORY so that condor_reconfig doesn't
        #   have access to the POOL signing key.
        # Override SEC_CLIENT_AUTHENTICATION_METHODS so that our special
        #   IDToken is used for authentication.
        # Override MASTER_ADDRESS_FILE so that condor_config_val has to
        #   query the collector to locate the master.
        rv = condor.run_command(
            ["env", "_condor_SEC_PASSWORD_DIRECTORY=/nothere",
                "_condor_SEC_CLIENT_AUTHENTICATION_METHODS=IDTOKEN",
                "_condor_MASTER_ADDRESS_FILE=/nothere",
                "condor_config_val", "-name", "test@here", "LOG"]
        )
        assert rv.returncode == 0



#!/usr/bin/python3

# Copyright 2020 HTCondor Team, Computer Sciences Department,
# University of Wisconsin-Madison, WI.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import enum
import re

try:
    from colorama import init
    init()
    HAS_COLORAMA = True
except ImportError:
    HAS_COLORAMA = False

IS_WINDOWS = os.name == "nt"
IS_TTY = sys.stdout.isatty()

ANSI_RESET = "\033[0m"
ANSI_ESCAPE_RE = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")

class Color(str, enum.Enum):
    BLACK = "\033[30m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"
    DARK_GREY = "\033[90m"
    BRIGHT_RED = "\033[91m"
    BRIGHT_GREEN = "\033[92m"
    BRIGHT_YELLOW = "\033[93m"
    BRIGHT_BLUE = "\033[94m"
    BRIGHT_MAGENTA = "\033[95m"
    BRIGHT_CYAN = "\033[96m"
    BRIGHT_WHITE = "\033[97m"

class Style(str, enum.Enum):
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    BLINK = "\033[5m"
    # ITALIC = "\033[3m" Note: Not widely supported

class AnsiOptions():
    def __init__(self, **kwargs):
        self.bold = kwargs.get("bold", False)
        self.blink = kwargs.get("blink", False)
        self.underline = kwargs.get("underline", False)
        self.style = kwargs.get("style")
        self.color = kwargs.get("color")
        self.color_id = kwargs.get("color_id")
        self.highlight = kwargs.get("highlight")

        if self.style is not None:
            self.bold = True if self.style == Style.BOLD else self.bold
            self.blink = True if self.style == Style.BLINK else self.blink
            self.underline = True if self.style == Style.UNDERLINE else self.underline

        if self.color is not None:
            self.color = self.color[2:-1]

        if self.color_id is not None:
            self.color_id = sorted([0, self.color_id, 255])[1]

        if self.highlight is not None:
            self.highlight = sorted([0, self.highlight, 255])[1]

    def __str__(self):
        style = "\033["
        style += "1;" if self.bold else ""
        style += "4;" if self.underline else ""
        style += "5;" if self.blink else ""
        style += f"38;5;{self.color_id};" if self.color_id is not None else (f"{self.color};" if self.color is not None else "")
        style += f"48;5;{self.highlight};" if self.highlight is not None else ""
        return f"{style[:-1]}m" if ";" in style else ""

def is_capable():
    return IS_TTY and (not IS_WINDOWS or HAS_COLORAMA)

def colorize(string: str, color: Color) -> str:
    return color + string + ANSI_RESET if is_capable() else string

# See https://gist.github.com/fnky/458719343aabd01cfb17a3a4f7296797#256-colors for values
def id_colorize(string: str, id: int) -> str:
    id = sorted([0, id, 255])[1]
    return f"\033[38;5;{id}m" + string + ANSI_RESET if is_capable() else string

# Uses same values as id_colorize()
def highlight(string: str, id: int) -> str:
    id = sorted([0, id, 255])[1]
    return f"\033[48;5;{id}m" + string + ANSI_RESET if is_capable() else string

def underline(string: str) -> str:
    return Style.UNDERLINE + string + ANSI_RESET if is_capable() else string

#def italicize(string: str) -> str:
#    return Style.ITALIC + string + ANSI_RESET if is_capable() else string

def bold(string: str) -> str:
    return Style.BOLD + string + ANSI_RESET if is_capable() else string

def blink(string: str) -> str:
    return Style.BLINK + string + ANSI_RESET if is_capable() else string

def strip_ansi(string: str) -> str:
    return ANSI_ESCAPE_RE.sub("", string)

def stylize(string: str, options: AnsiOptions) -> str:
    return str(options) + string + ANSI_RESET if is_capable() else string



from collections.abc import MutableMapping

from .htcondor2_impl import (
    _remote_param_keys,
    _remote_param_get,
    _remote_param_set,
)

from ._common_imports import (
    classad
)

class RemoteParam(MutableMapping):

    def __init__(self, location : classad.ClassAd):
        """
        The keys and values of this :class:`collections.abc.MutableMapping` are the keys
        and values of the specified daemon's configuration.  Assigning to
        a key sets the "runtime" configuration for the corresponding
        HTCondor macro, which won't take effect until the daemon has been
        reconfigured.  See :macro:`ENABLE_RUNTIME_CONFIG`.  Nonetheless,
        prior assignments will be reflected in subsequent lookups, but
        see :meth:`refresh`, below.

        :param location:  A ClassAd with a ``MyAddress`` attribute, such as
            might be returned by :meth:`htcondor2.Collector.locate`.
        """
        self._location = location
        self.refresh()


    def refresh(self):
        """
        Rebuild the dictionary based on the current configuration of the
        daemon.  Configuration values set by assigning to this dictionary
        do not become part of the current configuration until the daemon
        has been reconfigured; if you have not reconfigured the daemon,
        this method will result in a dictionary without those changes.
        """
        key_list = _remote_param_keys(self._location._handle)
        self._keys = key_list.split(',')
        self._cache = dict()


    def __getitem__(self, key : str):
        if not isinstance(key, str):
            raise TypeError("key must be a string")
        if not key in self._keys:
            raise KeyError(key)

        if not key in self._cache:
            self._cache[key] = _remote_param_get(self._location._handle, key)
        return self._cache[key]


    def __setitem__(self, key : str, value : str):
        if not isinstance(key, str):
            raise TypeError("key must be a string")
        if not isinstance(value, str):
            raise TypeError("value must be a string")

        if key not in self._keys:
            self._keys.append(key)
        self._cache[key] = value
        return _remote_param_set(self._location._handle, key, value)


    def __delitem__(self, key : str):
        if not isinstance(key, str):
            raise TypeError("key must be a string")
        if not key in self._keys:
            raise KeyError(key)

        del self._cache[key]
        self._keys.remove(key)
        return _remote_param_set(self._location._handle, key, "")


    def __iter__(self):
        for key in self._keys:
            yield(key)


    def __len__(self):
        return len(self._keys)




import sys
import enum


class CredType(enum.IntEnum):
    """
    An enumeration of the types of credentials that a *condor_credd*
    can manage.

    .. attribute:: Password

        On Windows, HTCondor can only impersonate a user whose password
        is in the *condor_credd*.

    .. attribute:: Kerberos

        HTCondor can manage Kerberos credentials, forwarding them to
        EPs so user jobs can access the user's files.

    .. attribute:: OAuth

        HTCondor can manage OAuth credentials, automatically renewing them
        and forwarding them to EPs, where they're typically used by file
        transfer plug-ins, although they are also available while the job
        is running.
    """

    Password = 0x24
    Kerberos = 0x20
    OAuth    = 0x28
